from cpython cimport bool
from .sqlfront cimport DBPROCESS, BYTE

cdef void log(char *, ...)

cdef struct _mssql_parameter_node:
    _mssql_parameter_node *next
    BYTE                  *value

cdef class MSSQLConnection:

    # class property variables
    cdef bint _connected
    cdef int _rows_affected
    cdef int _query_timeout
    cdef char *_charset
    cdef bool use_datetime2

    # class internal variables
    cdef DBPROCESS *dbproc
    cdef int last_msg_no
    cdef int last_msg_severity
    cdef int last_msg_state
    cdef int last_msg_line
    cdef int last_dbresults
    cdef int num_columns
    cdef public bint debug_queries
    cdef char *last_msg_str
    cdef char *last_msg_srv
    cdef char *last_msg_proc
    cdef tuple column_names
    cdef tuple column_types

    cdef object msghandler

    cpdef cancel(self)
    cdef void clear_metadata(self)
    cdef object convert_db_value(self, BYTE *, int, int)
    cdef int convert_python_value(self, object value, BYTE **, int*, int*) except -1
    cpdef execute_query(self, query, params=?)
    cpdef execute_non_query(self, query, params=?)
    cpdef execute_row(self, query, params=?)
    cpdef execute_scalar(self, query, params=?)
    cdef fetch_next_row(self, int, int)
    cdef format_and_run_query(self, query_string, params=?)
    cdef format_sql_command(self, format, params=?)
    cdef get_result(self)
    cdef get_row(self, int, int)

    cpdef set_msghandler(self, object handler)

    cdef bcp_init(self, object)
    cdef bcp_hint(self, BYTE * value, int valuelen)
    cdef bcp_bind(self, object value, int is_none, int column_db_type, int position, BYTE **data)
    cdef bcp_batch(self)
    cpdef bcp_sendrow(self, object element, object column_ids)
    cdef bcp_done(self)


cdef class MSSQLRowIterator:
    cdef MSSQLConnection conn
    cdef int row_format

cdef class MSSQLStoredProcedure:
    cdef MSSQLConnection conn
    cdef DBPROCESS *dbproc
    cdef char *procname
    cdef int param_count
    cdef bool had_positional
    cdef list output_indexes
    cdef dict params
    cdef _mssql_parameter_node *params_list




# -*- coding: utf-8 -*-
"""
Test connection to database.
"""

from __future__ import with_statement
from os import path, makedirs, environ
import re
import unittest
import tempfile

import pytest

from pymssql import _mssql

from .helpers import config, mssqlconn
server = config.server
username = config.user
password = config.password
database = config.database
port = config.port
ipaddress = config.ipaddress
instance = config.instance


@pytest.mark.mssql_server_required
class TestCons(unittest.TestCase):

    def connect(self, **kwargs):
        with tempfile.TemporaryDirectory() as tmpdir:
            config_dump_path = path.join(tmpdir, 'freetds-config-dump.txt')
            dump_path = path.join(tmpdir, 'freetds-dump.txt')
            environ['TDSDUMPCONFIG'] = config_dump_path
            environ['TDSDUMP'] = dump_path
            _mssql.connect(**kwargs)
            with open(config_dump_path, 'r') as fh:
                return fh.read()

    def test_connection_tds_version(self):

        for tds_version in ('4.2', '7.0', '7.1', '7.2', '7.3', '7.4'):
            with self.subTest(f"{tds_version}", tds_version=tds_version):
                cdump = self.connect(server=server, port=port,
                                    user=username, password=password,
                                    tds_version=tds_version)
                major_version = re.search('major_version = (\\S+)', cdump).groups()[0]
                minor_version = re.search('minor_version = (\\S+)', cdump).groups()[0]
                actual_version = f"{major_version}.{minor_version}"
                self.assertEqual(tds_version, actual_version)

    def test_connection_by_dns_name(self):
        cdump = self.connect(server=server, port=port, user=username, password=password)
        dump_server_name = re.search('server_name = (\\S+)', cdump).groups()[0]
        self.assertIn(server, dump_server_name)
        dump_server_host_name = re.search('server_host_name = (\\S+)', cdump).groups()[0]
        self.assertEqual(dump_server_host_name, server)
        dump_user_name = re.search('user_name = (\\S+)', cdump).groups()[0]
        self.assertEqual(dump_user_name, username)
        dump_port = re.search('port = (\\S+)', cdump).groups()[0]
        self.assertIn(port, dump_port)

    def test_connection_by_ip(self):
        cdump = self.connect(server=ipaddress, port=port, user=username, password=password)
        dump_server_name = re.search('server_name = (\\S+)', cdump).groups()[0]
        self.assertIn(ipaddress, dump_server_name)
        dump_server_host_name = re.search('server_host_name = (\\S+)', cdump).groups()[0]
        self.assertEqual(dump_server_host_name, ipaddress)

    def test_port_override_ipaddress(self):
        server_join = '%s:%s' % (ipaddress, port)
        cdump = self.connect(server=server_join, user=username, password=password)
        dump_server_name = re.search('server_name = (\\S+)', cdump).groups()[0]
        self.assertIn(ipaddress, dump_server_name)
        dump_server_host_name = re.search('server_host_name = (\\S+)', cdump).groups()[0]
        self.assertEqual(dump_server_host_name, ipaddress)
        dump_port = re.search('port = (\\S+)', cdump).groups()[0]
        self.assertIn(port, dump_port)

    def test_port_override_name(self):
        server_join = '%s:%s' % (server, port)
        cdump = self.connect(server=server_join, user=username, password=password)
        dump_server_name = re.search('server_name = (\\S+)', cdump).groups()[0]
        self.assertIn(server, dump_server_name)
        dump_server_host_name = re.search('server_host_name = (\\S+)', cdump).groups()[0]
        self.assertEqual(dump_server_host_name, server)
        dump_port = re.search('port = (\\S+)', cdump).groups()[0]
        self.assertIn(port, dump_port)

    def test_instance(self):
        if not instance:
            pytest.skip()
        server_join = r'%s\%s' % (server, instance)
        cdump = self.connect(server=server_join, user=username, password=password)
        dump_server_name = re.search('server_name = (\\S+)', cdump).groups()[0]
        self.assertIn(server, dump_server_name)
        dump_server_host_name = re.search('server_host_name = (\\S+)', cdump).groups()[0]
        self.assertEqual(dump_server_host_name, server)
        dump_port = re.search('port = (\\S+)', cdump).groups()[0]
        self.assertEqual(dump_port, 0)

    def test_valid_tds_version_property(self):
        # Issue #211 (https://github.com/pymssql/pymssql/issues/211)
        conn = mssqlconn()
        self.assertIsNotNone(conn.tds_version)
        self.assertTrue(conn.tds_version > 0)
        conn.close()

    def test_conn_props_override(self):
        conn = mssqlconn(conn_properties='SET TEXTSIZE 2147483647')
        conn.close()

        conn = mssqlconn(conn_properties='SET TEXTSIZE 2147483647;')
        conn.close()

        conn = mssqlconn(conn_properties='SET TEXTSIZE 2147483647;SET ANSI_NULLS ON;')
        conn.close()

        conn = mssqlconn(conn_properties='SET TEXTSIZE 2147483647;SET ANSI_NULLS ON')
        conn.close()

        conn = mssqlconn(conn_properties='SET TEXTSIZE 2147483647;'
                         'SET ANSI_NULLS ON;')
        conn.close()

        conn = mssqlconn(conn_properties=['SET TEXTSIZE 2147483647;', 'SET ANSI_NULLS ON'])
        conn.close()
        self.assertRaises(_mssql.MSSQLDriverException, mssqlconn, conn_properties='BOGUS SQL')

        conn = _mssql.connect(
            conn_properties='SET TEXTSIZE 2147483647',
            server=server,
            user=username,
            password=password
        )
        conn.close()


@pytest.mark.slow
class TestFailedConnection(unittest.TestCase):

    @pytest.mark.xfail(strict=False, reason="Could timeout, or fail with different error messages")
    @pytest.mark.timeout(600)
    def test_repeated_failed_connections(self):
        # This is a test for https://github.com/pymssql/pymssql/issues/145
        # (Repeated failed connections result in error string getting longer
        # and longer)

        _mssql.login_timeout = 5
        last_exc_message = None
        for i in range(5):
            try:
                _mssql.connect(
                    server='www.google.com',
                    port=81,
                    user='joe',
                    password='secret',
                    database='tempdb')
            except Exception as exc:
                exc_message = exc.args[0][1]

                if last_exc_message:
                    self.assertEqual(exc_message, last_exc_message)

                last_exc_message = exc_message



"""
This type stub file was generated by pyright.
"""

import datetime
import re
from collections.abc import Callable, Sequence
from typing import Any

class datetime2(datetime.datetime): ...

__full_version__: str
__version__: str
VERSION: tuple[int, ...]

class NoParams: ...

ROW_FORMAT_TUPLE: int
ROW_FORMAT_DICT: int
STRING: int
BINARY: int
NUMBER: int
DATETIME: int
DECIMAL: int
SQLBINARY: int
SQLBIT: int
SQLBITN: int
SQLCHAR: int
SQLDATETIME: int
SQLDATETIM4: int
SQLDATETIMN: int
SQLDECIMAL: int
SQLFLT4: int
SQLFLT8: int
SQLFLTN: int
SQLIMAGE: int
SQLINT1: int
SQLINT2: int
SQLINT4: int
SQLINT8: int
SQLINTN: int
SQLMONEY: int
SQLMONEY4: int
SQLMONEYN: int
SQLNUMERIC: int
SQLREAL: int
SQLTEXT: int
SQLVARBINARY: int
SQLVARCHAR: int
SQLUUID: int
SQLDATE: int
SQLTIME: int
SQLDATETIME2: int

def py2db_type(py_type: type, value: Any) -> int: ...

class MSSQLException(Exception):
    """
    Base exception class for the MSSQL driver.
    """


class MSSQLDriverException(MSSQLException):
    """
    Inherits from the base class and raised when an error is caused within
    the driver itself.
    """


class MSSQLDatabaseException(MSSQLException):
    """
    Raised when an error occurs within the database.
    """

    message: str

login_timeout: int
min_error_severity: int
wait_callback: None

def set_wait_callback(a_callable: Callable) -> None: ...

class MSSQLRowIterator:
    def __init__(self, connection: MSSQLConnection, row_format: int) -> None: ...
    def __iter__(self) -> MSSQLRowIterator: ...
    def __next__(self) -> tuple[Any, ...]: ...

class MSSQLConnection:
    charset: str
    connected: bool
    identity: tuple[Any, ...]
    query_timeout: int
    rows_affected: int
    tds_version: int
    tds_version_tuple: tuple[int, int] | None
    def __init__(
        self,
        server: str = ".",
        user: str | None = None,
        password: str | None = None,
        charset: str = "UTF-8",
        database: str = "",
        appname: str | None = None,
        port: str = "1433",
        tds_version: str | None = None,
        encryption: str | None = None,
        read_only: bool = False,
        use_datetime2=False,
        conn_properties: str | list[str] | None = None,
    ) -> None: ...
    def __dealloc__(self) -> None: ...
    def __enter__(self) -> MSSQLConnection: ...
    def __exit__(self, exc_type, exc_value, traceback) -> None: ...
    def __iter__(self) -> MSSQLRowIterator: ...
    def set_msghandler(self, handler: Callable) -> None:
        """
        set_msghandler(handler) -- set the msghandler for the connection

        This function allows setting a msghandler for the connection to
        allow a client to gain access to the messages returned from the
        server.
        """
        ...

    def cancel(self) -> None:
        """
        cancel() -- cancel all pending results.

        This function cancels all pending results from the last SQL operation.
        It can be called more than once in a row. No exception is raised in
        this case.
        """
        ...

    def close(self) -> None:
        """
        close() -- close connection to an MS SQL Server.

        This function tries to close the connection.  It can be called more than once in a row. No exception is raised
        in this case.
        """
        ...

    def mark_disconnected(self) -> None: ...
    def execute_non_query(self, query_string: str, params: object = ...) -> None:
        """
        execute_non_query(query_string, params=NoParams)

        This method sends a query to the MS SQL Server to which this object
        instance is connected. After completion, its results (if any) are
        discarded. An exception is raised on failure. If there are any pending
        results or rows prior to executing this command, they are silently
        discarded. This method accepts Python formatting. Please see
        execute_query() for more details.

        This method is useful for INSERT, UPDATE, DELETE and for Data
        Definition Language commands, i.e. when you need to alter your database
        schema.

        After calling this method, rows_affected property contains number of
        rows affected by the last SQL command.
        """
        ...

    def execute_query(self, query_string: str, params: object = ...) -> None:
        """
        execute_query(query_string, params=NoParams)

        This method sends a query to the MS SQL Server to which this object
        instance is connected. An exception is raised on failure. If there
        are pending results or rows prior to executing this command, they
        are silently discarded. After calling this method you may iterate
        over the connection object to get rows returned by the query.

        You can use Python formatting here and all values get properly
        quoted:
            conn.execute_query('SELECT * FROM empl WHERE id=%d', 13)
            conn.execute_query('SELECT * FROM empl WHERE id IN %s', ((5,6),))
            conn.execute_query('SELECT * FROM empl WHERE name=%s', 'John Doe')
            conn.execute_query('SELECT * FROM empl WHERE name LIKE %s', 'J%')
            conn.execute_query('SELECT * FROM empl WHERE name=%(name)s AND \
                city=%(city)s', { 'name': 'John Doe', 'city': 'Nowhere' } )
            conn.execute_query('SELECT * FROM cust WHERE salesrep=%s \
                AND id IN (%s)', ('John Doe', (1,2,3)))
            conn.execute_query('SELECT * FROM empl WHERE id IN %s',\
                (tuple(xrange(4)),))
            conn.execute_query('SELECT * FROM empl WHERE id IN %s',\
                (tuple([3,5,7,11]),))

        This method is intended to be used on queries that return results,
        i.e. SELECT. After calling this method AND reading all rows from,
        result rows_affected property contains number of rows returned by
        last command (this is how MS SQL returns it).
        """
        ...

    def execute_row(self, query_string: str, params: object = ...) -> tuple[Any, ...]:
        """
        execute_row(query_string, params=NoParams)

        This method sends a query to the MS SQL Server to which this object
        instance is connected, then returns first row of data from result.

        An exception is raised on failure. If there are pending results or
        rows prior to executing this command, they are silently discarded.

        This method accepts Python formatting. Please see execute_query()
        for details.

        This method is useful if you want just a single row and don't want
        or don't need to iterate, as in:

        conn.execute_row('SELECT * FROM employees WHERE id=%d', 13)

        This method works exactly the same as 'iter(conn).next()'. Remaining
        rows, if any, can still be iterated after calling this method.
        """
        ...

    def execute_scalar(
        self, query_string: str, params: object = ...
    ) -> tuple[Any, ...] | None:
        """
        execute_scalar(query_string, params=NoParams)

        This method sends a query to the MS SQL Server to which this object
        instance is connected, then returns first column of first row from
        result. An exception is raised on failure. If there are pending

        results or rows prior to executing this command, they are silently
        discarded.

        This method accepts Python formatting. Please see execute_query()
        for details.

        This method is useful if you want just a single value, as in:
            conn.execute_scalar('SELECT COUNT(*) FROM employees')

        This method works in the same way as 'iter(conn).next()[0]'.
        Remaining rows, if any, can still be iterated after calling this
        method.
        """
        ...

    def fetch_next_row(self, throw: int, row_format: int) -> tuple[Any, ...] | None: ...
    def format_and_run_query(self, query_string: str, params: object = ...) -> None:
        """
        This is a helper function, which does most of the work needed by any
        execute_*() function. It returns NULL on error, None on success.
        """
        ...

    def executemany(
        self, query_string: str, seq_of_parameters: Sequence[str], batch_size: int
    ) -> None: ...
    def get_header(self) -> tuple[str, ...] | None:
        """
        get_header() -- get the Python DB-API compliant header information.

        This method is infrastructure and doesn't need to be called by your
        code. It returns a list of 7-element tuples describing the current
        result header. Only name and DB-API compliant type is filled, rest
        of the data is None, as permitted by the specs.
        """
        ...

    def get_iterator(self, row_format: int) -> MSSQLRowIterator:
        """
        get_iterator(row_format) -- allows the format of the iterator to be specified

        While the iter(conn) call will always return a dictionary, this
        meth
        """
        ...

    def init_procedure(self, procname: str) -> MSSQLStoredProcedure:
        """
        init_procedure(procname) -- creates and returns a MSSQLStoredProcedure
        object.

        This methods initializes a stored procedure or function on the server
        and creates a MSSQLStoredProcedure object that allows parameters to
        be bound.
        """
        ...

    def nextresult(self) -> int | None:
        """
        nextresult() -- move to the next result, skipping all pending rows.

        This method fetches and discards any rows remaining from the current
        resultset, then it advances to the next (if any) resultset. Returns
        True if the next resultset is available, otherwise None.
        """
        ...

    def select_db(self, dbname: str) -> None:
        """
        select_db(dbname) -- Select the current database.

        This function selects the given database. An exception is raised on
        failure.
        """
        ...

    def bcp_sendrow(self, element: list[str], column_ids: list[int]): ...

class MSSQLStoredProcedure:
    connection: MSSQLConnection
    name: str
    parameters: list[str]
    def __init__(self, name: bytes, connection: MSSQLConnection) -> None: ...
    def __dealloc__(self) -> None: ...
    def bind(
        self,
        value: Any,
        dbtype: int,
        param_name: str | None = None,
        output: int | bool = False,
        null: int | bool = False,
        max_length: int = -1,
    ) -> None:
        """
        bind(value, data_type, param_name = None, output = False,
            null = False, max_length = -1) -- bind a parameter

        This method binds a parameter to the stored procedure.
        """
        ...

    def execute(self) -> None: ...

def remove_locale(value: bytes) -> str: ...

_re_pos_param: re.Pattern[bytes]
_re_name_param: re.Pattern[bytes]

def quote_simple_value(value: Any | None) -> str | bytes | None: ...
def quote_or_flatten(data: list[Any] | tuple[Any, ...]) -> str | bytes | None: ...
def quote_data(
    data: list[Any] | tuple[Any, ...] | dict[Any, Any],
) -> str | bytes | set[Any] | tuple[Any, ...] | None: ...
def substitute_params(toformat: str, params: object = ..., charset: str = "utf-8"): ...
def connect(*args, **kwargs) -> MSSQLConnection: ...

MssqlDatabaseException = MSSQLDatabaseException
MssqlDriverException = MSSQLDriverException
MssqlConnection = MSSQLConnection

def test_err_handler(
    connection: MSSQLConnection,
    severity: int,
    dberr: int,
    oserr: int,
    dberrstr: str,
    oserrstr: str,
) -> tuple[int, str, int, int, int]:
    """
    Expose err_handler function and its side effects to facilitate testing.
    """
    ...

def get_max_connections() -> int:
    """
    Get maximum simultaneous connections db-lib will open to the server.
    """
    ...

def set_max_connections(limit: int) -> None:
    """
    Set maximum simultaneous connections db-lib will open to the server.

    :param limit: the connection limit
    :type limit: int
    """
    ...

def get_dbversion() -> str:
    """
    Return string representing the version of db-lib.
    """
    ...



# -*- coding: utf-8 -*-
"""
Pytest configuration.
"""

import decimal
import os
from configparser import ConfigParser

import pytest

import tests.helpers as th
from .helpers import clear_db, get_app_lock, release_app_lock

cdir = os.path.dirname(__file__)
cfgpath = os.path.join(cdir, 'tests.cfg')

@pytest.fixture(scope="module")
def mssql_conn():
    return th.mssqlconn()

_parser = ConfigParser({
    'server': 'localhost',
    'username': 'sa',
    'password': 'sqlServerPassw0rd',
    'database': 'tempdb',
    'port': '1433',
    'ipaddress': '127.0.0.1',
    'instance': '',
})

optional_markers = {
    "slow": {"help": "Skip long tests",
             "marker-descr": "Mark tests that run longer than ~3 seconds",
             "skip-reason": "Test runs too long."},
    "mssql_server_required": {"help": "Skip tests that require MSSQL server",
             "marker-descr": "Mark tests that require MSSQL server",
             "skip-reason": "Test only runs if MSSQL server is available."},
    # add further markers here
}

def pytest_addoption(parser):
    parser.addoption(
        "--pymssql-section",
        type=str,
        default=os.environ.get('PYMSSQL_TEST_CONFIG', None),
        help="The name of the section to use from tests.cfg"
    )
    for marker, info in optional_markers.items():
        parser.addoption("--skip-{}".format(marker.replace('_','-')), action="store_true",
                         default=False, help=info['help'])

def pytest_configure(config):
    section = config.getoption('--pymssql-section')
    if section is not None:
        if not os.path.exists(cfgpath):
            raise ValueError(f"Config file '{cfgpath}' dos not exist.")
        _parser.read(cfgpath)
        if not _parser.has_section(section):
            raise ValueError('the tests.cfg file does not have section: %s' % section)
    else:
        section = 'DEFAULT'

    th.config.server = os.getenv('PYMSSQL_TEST_SERVER') or _parser.get(section, 'server')
    th.config.user = os.getenv('PYMSSQL_TEST_USERNAME') or _parser.get(section, 'username')
    th.config.password = os.getenv('PYMSSQL_TEST_PASSWORD') or _parser.get(section, 'password')
    th.config.database = os.getenv('PYMSSQL_TEST_DATABASE') or _parser.get(section, 'database')
    th.config.port = os.getenv('PYMSSQL_TEST_PORT') or _parser.get(section, 'port')
    th.config.ipaddress = os.getenv('PYMSSQL_TEST_IPADDRESS') or _parser.get(section, 'ipaddress')
    th.config.instance = os.getenv('PYMSSQL_TEST_INSTANCE') or _parser.get(section, 'instance')
    th.config.orig_decimal_prec = decimal.getcontext().prec

    for marker, info in optional_markers.items():
        config.addinivalue_line("markers",
                                "{}: {}".format(marker, info['marker-descr']))

    if get_app_lock():
        clear_db()

def pytest_unconfigure(config):
    release_app_lock()


def pytest_collection_modifyitems(config, items):

    marker = "mssql_server_required"
    info = optional_markers[marker]
    if th.global_mssqlconn is None or config.getoption("--skip-{}".format(marker.replace('_','-'))):
        skip = pytest.mark.skip(reason=info['skip-reason'])
        for item in items:
            if marker in item.keywords:
                item.add_marker(skip)
    marker = "slow"
    info = optional_markers[marker]
    if th.global_mssqlconn is None or config.getoption("--skip-{}".format(marker)):
        skip = pytest.mark.skip(reason=info['skip-reason'])
        for item in items:
            if marker in item.keywords:
                item.add_marker(skip)



# -*- coding: utf-8 -*-
"""
Test context managers -- i.e.: the `with` statement
"""

import unittest

import pytest

from pymssql import InterfaceError
from .helpers import pymssqlconn, mssqlconn


@pytest.mark.mssql_server_required
class TestContextManagers(unittest.TestCase):
    def test_pymssql_Connection_with(self):
        with pymssqlconn() as conn:
            cursor = conn.cursor()
            cursor.execute("SELECT @@version AS version")
            self.assertIsNotNone(conn._conn)

        with self.assertRaises(InterfaceError) as context:
            self.assertIsNotNone(conn._conn)

        self.assertEqual(str(context.exception), "Connection is closed.")

    def test_pymssql_Cursor_with(self):
        conn = pymssqlconn()
        with conn.cursor() as cursor:
            cursor.execute("SELECT @@version AS version")
            self.assertIsNotNone(conn._conn)

        self.assertIsNotNone(cursor)

        with self.assertRaises(InterfaceError) as context:
            cursor.execute("SELECT @@version AS version")

        self.assertEqual(str(context.exception), "Cursor is closed.")

    def test_mssql_Connection_with(self):
        with mssqlconn() as conn:
            conn.execute_query("SELECT @@version AS version")
            self.assertTrue(conn.connected)

        self.assertFalse(conn.connected)



# -*- coding: utf-8 -*-
"""
Test connection config.
"""

from __future__ import with_statement
from os import path, environ
import tempfile

import pytest

from pymssql import _mssql

@pytest.mark.mssql_server_required
class TestConfig(object):
    def connect(self, **kwargs):
        with tempfile.TemporaryDirectory() as tmpdir:
            config_dump_path = path.join(tmpdir, 'freetds-config-dump.txt')
            environ['TDSDUMPCONFIG'] = config_dump_path
            try:
                _mssql.connect(**kwargs)
                assert False
            except _mssql.MSSQLDriverException as e:
                # we get this when the name of the server is not valid
                if 'Connection to the database failed' not in str(e):
                    raise
            except _mssql.MSSQLDatabaseException as e:
                # we get this when the name or IP can be obtained but the connection
                # can not be made
                if e.args[0][0] != 20009:
                    raise
            with open(config_dump_path, 'r') as fh:
                return fh.read()

    @pytest.mark.slow
    def test_config_values(self):
        config_dump = self.connect(
            server='dontnameyourserverthis',
            user = 'bob',
            database = 'tempdb',
            tds_version='7.1'
            )
        assert 'user_name = bob' in config_dump
        assert 'database = tempdb\n' in config_dump
        # test default port
        assert 'port = 1433' in config_dump
        # not sure why 7.1 version is used instead of 8.0 which is the
        # default
        assert 'major_version = 7' in config_dump
        assert 'minor_version = 1' in config_dump

    @pytest.mark.slow
    def test_tds_protocol_version_42(self):
        config_dump = self.connect(
            server='dontnameyourserverthis',
            tds_version='4.2'
            )
        assert 'major_version = 4' in config_dump
        assert 'minor_version = 2' in config_dump

    @pytest.mark.slow
    def test_tds_protocol_version_70(self):
        config_dump = self.connect(
            server='dontnameyourserverthis',
            tds_version='7.0'
            )
        assert 'major_version = 7' in config_dump
        assert 'minor_version = 0' in config_dump

    @pytest.mark.slow
    def test_tds_protocol_version_71(self):
        config_dump = self.connect(
            server='dontnameyourserverthis',
            tds_version='7.1'
            )
        assert 'major_version = 7' in config_dump
        assert 'minor_version = 1' in config_dump

    @pytest.mark.slow
    def test_tds_protocol_version_80(self):
        # follow-up: turns out 8.0 was erroneous.  MS named the new protocol
        # 7.1 instead of 8.0, so FreeTDS will accept 8.0 but shows as 7.1.
        # got that from the FreeTDS mailing list.  New FreeTDS docs,built from
        # source, have a page that describes the protocol and that page lists
        # versions 7.0, 7.1, and 7.2 among others.

        config_dump = self.connect(
            server='dontnameyourserverthis',
            tds_version='8.0'
            )
        assert 'major_version = 7' in config_dump
        assert 'minor_version = 1' in config_dump

    @pytest.mark.slow
    def test_tds_protocol_version_72(self):
        config_dump = self.connect(
            server='dontnameyourserverthis',
            tds_version='7.2'
            )
        assert 'major_version = 7' in config_dump
        assert 'minor_version = 2' in config_dump

    def test_tds_protocol_version_invalid(self):
        try:
            self.connect(tds_version='1.0')
            assert False
        except _mssql.MSSQLException as e:
            assert 'unrecognized tds version: 1.0' == str(e)

    def test_tds_nonstandard_port_int(self):
        #it should convert it to a string
        config_dump = self.connect(port=1435)
        assert 'port = 1435' in config_dump



# -*- coding: utf-8 -*-

import gc
import sys
import psutil
import pymssql
import pytest


@pytest.mark.slow
@pytest.mark.xfail(reason="Memory test is not stable")
def test_memory_leak_on_unsuccessful_connect():
    """
    This test checks python process memory usage and not just unsuccessful
    connect path. Many other factors (i.e. garbage collection) affect memory
    usage pattern. So this is an indirect test which is somewhat flaky and
    because of that is marked 'xfail'.
    """
    p = psutil.Process()
    m0 = p.memory_full_info()
    for i in range(10):
        gc.collect()
        try:
            pymssql.connect(server="www.google.com", port=81, user='username',
                            password='password', login_timeout=1)
        except:
            pass
        gc.collect()
        m1 = p.memory_full_info()
        duss = m1.uss - m0.uss
        print(i, "uss=", m1.uss, "duss:", duss)
        if i > 5:
            assert duss <= 0
        m0 = m1


if __name__ == '__main__':

    test_memory_leak_on_unsuccessful_connect()



# -*- coding: utf-8 -*-
"""
Test bulk copy.
"""

import unittest
import datetime

import pytest

from pymssql import _mssql
from tests.helpers import drop_table, pymssqlconn


tablename = "pymssql"
simple_table = "CREATE TABLE %s (a1 INT, a2 INT, a3 INT)" % tablename
complex_table = """
    CREATE TABLE %s (
        pk_id int IDENTITY (1, 1) NOT NULL,
        uuid uniqueidentifier DEFAULT newsequentialid(),
        col_real real UNIQUE,
        col_float float,
        col_datetime datetime,
        col_bit bit,
        col_varchar varchar(50),
        col_varbinary varbinary(50)
    )
""" % tablename


@pytest.mark.mssql_server_required
class TestTypes(unittest.TestCase):
    def setUp(self):
        self.conn = pymssqlconn()
        drop_table(self.conn._conn, tablename)

    def tearDown(self):
        self.conn.close()

    def expect_simple_table_content(self, query, content):
        self.conn._conn.execute_query(query)
        assert [(row[0], row[1], row[2]) for row in self.conn._conn] == content

    def expect_row_count(self, expected_row_count):
        self.conn._conn.execute_query('select count(*) from pymssql')
        assert tuple(self.conn._conn)[0][0] == expected_row_count

    def simple_table_test(self, content, **kwargs):
        self.conn._conn.execute_non_query(simple_table)
        self.conn.bulk_copy(tablename, content, **kwargs)
        self.expect_simple_table_content('select * from pymssql', content)

    def test_simple_table_bulk_copy(self):
        self.simple_table_test([(1, 2, 3), (4, 5, 6)])

    def test_lots_of_rows_single_batch(self):
        self.conn._conn.execute_non_query(simple_table)
        self.conn.bulk_copy(tablename, [(1, 2, 3), (4, 5, 6)] * 100000, batch_size=1000000)
        self.expect_simple_table_content('select top 2 * from pymssql', [(1, 2, 3), (4, 5, 6)])
        self.expect_row_count(200000)

    def test_batches(self):
        self.conn._conn.execute_non_query(simple_table)

        self.conn.bulk_copy(tablename, [(1, 2, 3), (4, 5, 6)] * 100000, batch_size=1000)

        self.expect_simple_table_content('select top 2 * from pymssql', [(1, 2, 3), (4, 5, 6)])
        self.expect_row_count(200000)

    def test_exact_batch_size(self):
        self.conn._conn.execute_non_query(simple_table)

        self.conn.bulk_copy(tablename, [(1, 2, 3), (4, 5, 6)] * 500, batch_size=1000)

        self.expect_simple_table_content('select top 2 * from pymssql', [(1, 2, 3), (4, 5, 6)])
        self.expect_row_count(1000)

    def test_tablock_hint(self):
        self.simple_table_test([(1, 2, 3), (4, 5, 6)], tablock=True)

    def test_check_constraints_hint(self):
        self.simple_table_test([(1, 2, 3), (4, 5, 6)], check_constraints=True)

    def test_fire_triggers_hint(self):
        self.simple_table_test([(1, 2, 3), (4, 5, 6)], fire_triggers=True)

    def test_null_values(self):
        self.simple_table_test([(1, None, 3), (None, None, None), (1, 2, 3)])

    def test_column_ids(self):
        self.conn._conn.execute_non_query(simple_table)
        self.conn.bulk_copy(tablename, [(1, 2, 3), (4, 5, 6)], column_ids=[1, 3, 2])
        self.expect_simple_table_content('select * from pymssql', [(1, 3, 2), (4, 6, 5)])

    def test_too_many_columns(self):
        self.conn._conn.execute_non_query(simple_table)
        with self.assertRaises(_mssql.MSSQLDatabaseException):
            self.conn.bulk_copy(tablename, [(7, 7, 7, 7)])

    def test_bad_value(self):
        self.conn._conn.execute_non_query(simple_table)
        with self.assertRaises(_mssql.MSSQLDatabaseException):
            self.conn.bulk_copy(tablename, [("Hello", 7, 7)])

    def test_too_few_column_ids(self):
        self.conn._conn.execute_non_query(simple_table)
        caught_exception = False

        try:
            self.conn.bulk_copy(tablename, [(1, 2, 3)], column_ids=[1])
        except ValueError:
            caught_exception = True

        assert caught_exception

    def test_invalid_column_ids(self):
        self.conn._conn.execute_non_query(simple_table)
        with self.assertRaises(_mssql.MSSQLDatabaseException):
            self.conn.bulk_copy(tablename, [(1, 2, 3)], column_ids=[1, 2, 9])

    def test_complex_table(self):
        self.conn._conn.execute_non_query(complex_table)
        rows = [
            (1.2000000476837158, 3.4, datetime.datetime(year=2020, month=1, day=2, hour=3, minute=4, second=5), True, "Hello World", b'\x02\x03\x05\x07'),
            (5.599999904632568, 7.8, datetime.datetime(year=2021, month=2, day=3, hour=4, minute=5, second=6), False, "Hello World!", bytearray([2, 3, 5, 7])),
        ]
        self.conn.bulk_copy(tablename, rows, [3, 4, 5, 6, 7, 8])
        self.conn._conn.execute_query('select * from pymssql')
        assert [tuple(row[i] for i in range(2, 8)) for row in self.conn._conn] == rows

    def test_uniqueness_failure(self):
        self.conn._conn.execute_non_query(complex_table)

        rows = [
            (1.2000000476837158, 3.4, datetime.datetime(year=2020, month=1, day=2, hour=3, minute=4, second=5), True, "Hello World"),
            (1.2000000476837158, 7.8, datetime.datetime(year=2021, month=2, day=3, hour=4, minute=5, second=6), False, "Hello World!"),
        ]
        with self.assertRaises(_mssql.MSSQLDatabaseException):
            self.conn.bulk_copy(tablename, rows, [3, 4, 5, 6, 7])



# -*- coding: utf-8 -*-
"""
Test connection with as_dict=True.
"""

import unittest

import pytest

import pymssql

from .helpers import config

def pymssqlconn(**kwargs):
    return pymssql.connect(
            server=config.server,
            user=config.user,
            password=config.password,
            database=config.database,
            port=config.port,
            **kwargs
        )


@pytest.mark.mssql_server_required
class TestConnectionAsDict(unittest.TestCase):

    def setUp(self):
        self.conn = pymssqlconn(as_dict=True)

    def test_fetchall_with_connection_as_dict(self):
        # This test is for http://code.google.com/p/pymssql/issues/detail?id=18
        cursor = self.conn.cursor()
        cursor.execute("SELECT 'foo' AS first_name, 'bar' AS last_name")
        data = cursor.fetchall()
        self.assertEqual(data, [{'first_name': 'foo', 'last_name': 'bar'}])

    def test_no_results_with_connection_as_dict(self):
        # Make sure that checking for columns without names doesn't break
        # statements that don't return results

        cursor = self.conn.cursor()
        cursor.execute("""
        CREATE TABLE daily_measurement (
            datetime DATETIME,
            value FLOAT,
            notes VARCHAR,
        )
        """)




# -*- coding: utf-8 -*-

import datetime
import unittest

from pymssql import datetime2
from .helpers import TestCaseWithTable, mssqlconn, get_sql_server_version

import pytest


class TestCaseDATETIME2(unittest.TestCase, TestCaseWithTable):

    @classmethod
    def setup_class(cls):
        cls.conn = mssqlconn()
        if get_sql_server_version(cls.conn) < 2008:
            pytest.skip("DATETIME2 field type isn't supported by SQL Server versions prior to 2008.")
        if cls.conn.tds_version < 7.3:
            pytest.skip("DATETIME2 field type isn't supported by TDS protocol older than 7.3.")
        cls.create_table()


class Test_DATETIME2(TestCaseDATETIME2):

    table_name = "test_datetime2"
    ddl_create = f"CREATE TABLE {table_name} (test DATETIME2)"
    min_date = datetime2(1, 1, 1, 0, 0, 0, 0)
    max_date = datetime2(9999, 12, 31, 23, 59, 59, 999999)

    def test_min_select(self):
        self.conn.execute_query("SELECT CAST ('0001-1-1 0:0:0' as DATETIME2)")
        res = tuple(self.conn)[0][0]
        assert isinstance(res, datetime.datetime)
        assert res == self.min_date

    def test_min_insert(self):
        res = self.insert_and_select('test', self.min_date)
        assert isinstance(res, datetime.datetime)
        assert res == self.min_date

    def test_max_select(self):
        self.conn.execute_query("SELECT CAST ('9999-12-31 23:59:59.9999999' as DATETIME2)")
        res = tuple(self.conn)[0][0]
        assert isinstance(res, datetime.datetime)
        assert res == self.max_date

    def test_max_insert(self):
        res = self.insert_and_select('test', self.max_date)
        assert isinstance(res, datetime.datetime)
        assert res == self.max_date

    def test_datetime2(self):
        testval = datetime2(2013, 1, 2, 3, 4, 5, 6)
        res = self.insert_and_select('test', testval)
        assert isinstance(res, datetime.datetime)
        assert testval == res

    def test_truncate(self):
        self.conn.execute_query("SELECT CAST ('2024-1-1 0:0:0.1234567' as DATETIME2)")
        res = tuple(self.conn)[0][0]
        assert isinstance(res, datetime.datetime)
        assert res == datetime.datetime(2024, 1, 1, 0, 0, 0, 123456)


class Test_695(TestCaseDATETIME2):
    """
    See issue #695
    """
    table_name = "dbo.test695"
    ddl_create = f"""
            CREATE TABLE {table_name} (
                valid_from DATETIME2 GENERATED ALWAYS AS ROW START NOT NULL,
                valid_to DATETIME2 GENERATED ALWAYS AS ROW END NOT NULL,
                PERIOD FOR SYSTEM_TIME (valid_from,valid_to),
                id INTEGER NOT NULL IDENTITY(1,1) PRIMARY KEY,
                test VARCHAR(255) NOT NULL,
                )
        """
        #WITH (SYSTEM_VERSIONING = ON (HISTORY_TABLE = dbo.test_history));

    def test_695(self):
        self.conn.execute_non_query(f"INSERT INTO {self.table_name} (test) VALUES (%s)", ("TEST", ))
        self.conn.execute_query(f"SELECT valid_from, valid_to FROM {self.table_name}")
        rows = tuple(self.conn)
        res = rows[0][1]
        assert isinstance(res, datetime.datetime)
        #print(rows)


class Test_DATETIME(TestCaseDATETIME2):

    table_name = "test_datetime"
    ddl_create = f"CREATE TABLE {table_name} (test DATETIME)"
    min_date = datetime.datetime(1753, 1, 1, 0, 0, 0, 0)
    max_date = datetime.datetime(9999, 12, 31, 23, 59, 59, 997000)

    def test_min_select(self):
        self.conn.execute_query("SELECT CAST ('1753-1-1 0:0:0' as DATETIME)")
        res = tuple(self.conn)[0][0]
        assert isinstance(res, datetime.datetime)
        assert not isinstance(res, datetime2)
        assert res == self.min_date

    def test_min_insert(self):
        res = self.insert_and_select('test', self.min_date)
        assert isinstance(res, datetime.datetime)
        assert not isinstance(res, datetime2)
        assert res == self.min_date

    def test_max_select(self):
        self.conn.execute_query("SELECT CAST ('9999-12-31 23:59:59.997' as DATETIME)")
        res = tuple(self.conn)[0][0]
        assert isinstance(res, datetime.datetime)
        assert not isinstance(res, datetime2)
        assert res == self.max_date

    def test_max_insert(self):
        res = self.insert_and_select('test', self.max_date)
        assert isinstance(res, datetime.datetime)
        assert not isinstance(res, datetime2)
        assert res == self.max_date

    def test_datetime(self):
        """
        Test for issue at https://code.google.com/p/pymssql/issues/detail?id=118
        datetime values are rounded to increments of .000, .003, or .007 seconds,
        see https://docs.microsoft.com/en-us/sql/t-sql/data-types/datetime-transact-sql
        """
        for mks in (0, 3000, 7000):
            testval = datetime.datetime(2013, 1, 2, 3, 4, 5, mks)
            self.conn.execute_non_query(f"DELETE FROM {self.table_name}")
            res = self.insert_and_select('test', testval)
            assert isinstance(res, datetime.datetime)
            assert not isinstance(res, datetime2)
            assert res == testval

    def test_datetime_params_as_dict(self):
        testval = datetime.datetime(2013, 1, 2, 3, 4, 5, 3000)
        res = self.insert_and_select('test', testval, params_as_dict=True)
        assert isinstance(res, datetime.datetime)
        assert not isinstance(res, datetime2)
        assert res == testval


class Test_DATE(TestCaseDATETIME2):

    table_name = "test_date"
    ddl_create = f"CREATE TABLE {table_name} (test DATE)"
    min_date = datetime.date(1, 1, 1)
    max_date = datetime.date(9999, 12, 31)

    def test_min_select(self):
        self.conn.execute_query("SELECT CAST ('0001-1-1' as DATE)")
        res = tuple(self.conn)[0][0]
        assert isinstance(res, datetime.date)
        assert res == self.min_date

    def test_min_insert(self):
        res = self.insert_and_select('test', self.min_date)
        assert isinstance(res, datetime.date)
        assert res == self.min_date

    def test_max_select(self):
        self.conn.execute_query("SELECT CAST ('9999-12-31' as DATE)")
        res = tuple(self.conn)[0][0]
        assert isinstance(res, datetime.date)
        assert res == self.max_date

    def test_max_insert(self):
        res = self.insert_and_select('test', self.max_date)
        assert isinstance(res, datetime.date)
        assert res == self.max_date

    def test_date(self):
        testval = datetime.date(2013, 1, 2)
        res = self.insert_and_select('test', testval)
        assert isinstance(res, datetime.date)
        assert res == testval

    def test_ancient_date(self):
        testval = datetime.date(13, 1, 2)
        res = self.insert_and_select('test', testval)
        assert isinstance(res, datetime.date)
        assert res == testval


class Test_TIME(TestCaseDATETIME2):

    table_name = "test_time"
    ddl_create = f"CREATE TABLE {table_name} (test TIME)"
    min_time = datetime.time(0, 0, 0, 0)
    max_time = datetime.time(23, 59, 59, 999999)

    def test_min_select(self):
        self.conn.execute_query("SELECT CAST ('0:0:0.0' as TIME)")
        res = tuple(self.conn)[0][0]
        assert isinstance(res, datetime.time)
        assert res == self.min_time

    def test_min_insert(self):
        res = self.insert_and_select('test', self.min_time)
        assert isinstance(res, datetime.time)
        assert res == self.min_time

    def test_max_select(self):
        self.conn.execute_query("SELECT CAST ('23:59:59.9999999' as TIME)")
        res = tuple(self.conn)[0][0]
        assert isinstance(res, datetime.time)
        assert res == self.max_time

    def test_max_insert(self):
        res = self.insert_and_select('test', self.max_time)
        assert isinstance(res, datetime.time)
        assert res == self.max_time

    def test_time(self):
        testval = datetime.time(3, 4, 5, 3000)
        res = self.insert_and_select('test', testval)
        assert isinstance(res, datetime.time)
        assert res == testval


class Test_DATETIMEOFFSET(TestCaseDATETIME2):

    table_name = "test_time"
    ddl_create = f"""CREATE TABLE {table_name} (
                        id INT default 1,
                        DateCreated DATETIMEOFFSET default sysdatetimeoffset()
                    )"""

    def test_649(self):
        """
        Test #649
        """
        self.conn.execute_non_query(
            f"INSERT INTO {self.table_name} (id) VALUES (%s)", (2, ))
        self.conn.execute_query(
            f'select DateCreated from {self.table_name} where id = 2')
        res = tuple(self.conn)[0][0]
        assert isinstance(res, datetime.datetime)
        assert res.strftime('%z') != ''

    def test_select_cast_0(self):
        self.conn.execute_query(
            "SELECT CAST ('2019-06-20 09:54:40.09550' as DATETIMEOFFSET)")
        res = tuple(self.conn)[0][0]
        assert isinstance(res, datetime.datetime)
        assert res == datetime.datetime(2019, 6, 20, 9, 54, 40, 95500,
                                        tzinfo=datetime.timezone.utc)

    def test_select_cast(self):
        self.conn.execute_query(
            "SELECT CAST ('2019-06-20 09:54:40.09550 +04:00' as DATETIMEOFFSET)")
        res = tuple(self.conn)[0][0]
        assert isinstance(res, datetime.datetime)
        assert res == datetime.datetime(2019, 6, 20, 9, 54, 40, 95500,
                    tzinfo=datetime.timezone(datetime.timedelta(seconds=4*60*60)))

    def test_insert_select(self):
        testval = datetime.datetime(3, 4, 5, 3,
                    tzinfo=datetime.timezone(datetime.timedelta(seconds=4*60*60)))
        self.conn.execute_non_query(
            f"INSERT INTO {self.table_name} (id, DateCreated) VALUES (%s, %s)",
            (22, testval))
        self.conn.execute_query(
            f'select DateCreated from {self.table_name} where id = 22')
        res = tuple(self.conn)[0][0]
        assert isinstance(res, datetime.datetime)
        assert res.strftime('%z') != ''
        assert res == testval



"""
Test parameters substitution.
"""

import datetime
import decimal

import pytest

from pymssql._mssql import quote_simple_value, datetime2


def test_none():
    res = quote_simple_value(None)
    assert res == b'NULL'


@pytest.mark.parametrize('val', (0, 13))
def test_int(val):
    res = quote_simple_value(val)
    assert res == f"{val}".encode('utf-8')


@pytest.mark.parametrize('val', (0.0, 3.14))
def test_float(val):
    res = quote_simple_value(val)
    assert res == f"{val}".encode('utf-8')


@pytest.mark.parametrize('val', (True, False))
def test_bool(val):
    res = quote_simple_value(val)
    assert res == f"{int(val)}".encode('utf-8')


@pytest.mark.parametrize('val', (0.123, 13.789))
def test_decimal(val):
    res = quote_simple_value(decimal.Decimal(val))
    assert res == f"{decimal.Decimal(val)}".encode('utf-8')


def test_uuid():
    u = '3cde0c92-2674-481f-8fd4-feacf98d8504'
    res = quote_simple_value(u)
    assert res == f"N'{u}'".encode('utf-8')


@pytest.mark.parametrize('val', ('Фрязино', 'Здравствуй Мир'))
def test_str(val):
    res = quote_simple_value(val)
    assert res == f"N'{val}'".encode('utf-8')


@pytest.mark.parametrize('val', (b"Hello", b"Hello 'World!'"))
def test_bytes(val):
    res = quote_simple_value(val)
    assert res == b"'%s'"%(val.replace(b"'", b"''"))


@pytest.mark.parametrize('val', (
                         datetime.date(1, 2, 3),
                         datetime.date(2024, 3, 24)))
def test_date(val):
    res = quote_simple_value(val)
    assert res == f"'{val.year:04}-{val.month:02d}-{val.day:02d}'".encode()

@pytest.mark.parametrize('val', (
                         datetime.time(0, 0, 0, 1),
                         datetime.time(0, 0, 1),
                         datetime.time(1, 2, 59),
                         datetime.time(23, 3, 24)))
def test_time(val):
    res = quote_simple_value(val)
    assert res == f"'{val.hour:02d}:{val.minute:02d}:{val.second:02d}" \
                  f".{val.microsecond:06}'".encode()


@pytest.mark.parametrize('val', (
                         datetime.datetime(1, 2, 3, 23, 3, 59),
                         datetime.datetime(2024, 3, 24, 0, 0, 1)))
def test_datetime(val):
    res = quote_simple_value(val)
    assert res == f"'{val.year:04}-{val.month:02d}-{val.day:02d}" \
                  f" {val.hour:02d}:{val.minute:02d}:{val.second:02d}" \
                  f".{(val.microsecond // 1000):03d}'".encode()


@pytest.mark.parametrize('val', (
                         datetime.datetime(1, 2, 3, 23, 3, 59),
                         datetime.datetime(2024, 3, 24, 0, 0, 1)))
def test_datetime_use_datetime2(val):
    res = quote_simple_value(val, use_datetime2=True)
    assert res == f"'{val.year:04}-{val.month:02d}-{val.day:02d}" \
                  f" {val.hour:02d}:{val.minute:02d}:{val.second:02d}" \
                  f".{val.microsecond:06}'".encode()

@pytest.mark.parametrize('val', (
                         datetime2(1, 2, 3, 23, 3, 59),
                         datetime2(2024, 3, 24, 0, 0, 1)))
def test_datetime2(val):
    res = quote_simple_value(val)
    assert res == f"'{val.year:04}-{val.month:02d}-{val.day:02d}" \
                  f" {val.hour:02d}:{val.minute:02d}:{val.second:02d}" \
                  f".{val.microsecond:06}'".encode()

@pytest.mark.parametrize('val', (
                         datetime.datetime(1, 2, 3, 23, 3, 59,
                                           tzinfo=datetime.timezone.utc),
                         datetime.datetime(2024, 3, 24, 0, 0, 1,
                                           tzinfo=datetime.timezone.utc)))
def test_datetime_utc(val):
    res = quote_simple_value(val)
    assert res == f"'{val.year:04}-{val.month:02d}-{val.day:02d}" \
                  f" {val.hour:02d}:{val.minute:02d}:{val.second:02d}" \
                  f".{(val.microsecond // 1000):03d}'".encode()


@pytest.mark.parametrize('val', (
                         datetime.datetime(1, 2, 3, 23, 3, 59,
                                           tzinfo=datetime.timezone(
                                               datetime.timedelta(hours=3))),
                         datetime.datetime(2024, 3, 24, 0, 0, 1,
                                           tzinfo=datetime.timezone(
                                               datetime.timedelta(hours=3)))))
def test_datetime_tz(val):
    res = quote_simple_value(val)
    assert res == f"'{val.year:04}-{val.month:02d}-{val.day:02d}" \
                  f" {val.hour:02d}:{val.minute:02d}:{val.second:02d}" \
                  f".{val.microsecond:06}{val.strftime('%Z')[3:]}'".encode()



# Since Cython needs to know what you are using from header files this
# definition is provided so it knows exactly what we are using from
# FreeTDS.

cdef extern from "sqlfront.h":

    ## Type Definitions ##
    cdef struct tds_dblib_dbprocess:
        pass
    cdef struct tds_sysdep_real32_type:
        pass
    cdef struct tds_sysdep_real64_type:
        pass

    ctypedef tds_dblib_dbprocess DBPROCESS
    cdef struct tds_dblib_loginrec:
        pass
    ctypedef tds_dblib_loginrec LOGINREC
    ctypedef void DBCURSOR
    ctypedef int BOOL
    ctypedef short int SHORT
    ctypedef unsigned char BYTE
    ctypedef int RETCODE
    ctypedef short unsigned int DBUSMALLINT

    ctypedef unsigned char DBBINARY
    ctypedef int           DBBIT
    ctypedef unsigned char DBBOOL
    ctypedef char          DBCHAR
    ctypedef int           DBINT
    ctypedef tds_sysdep_real32_type DBREAL
    ctypedef tds_sysdep_real64_type DBFLT8
    cdef struct            DBMONEY:
        DBINT mnyhigh
        unsigned int mnylow
    cdef struct            DBMONEY4:
        DBINT mny4
    ctypedef unsigned char DBTINYINT
    ctypedef short int     DBSMALLINT

    ctypedef struct        DBDATETIME:
        DBINT dtdays
        DBINT dttime
    ctypedef struct        DBDATETIME4:
        DBUSMALLINT days
        DBUSMALLINT minutes

    ctypedef struct        DBCOL:
        DBINT SizeOfStruct
        DBCHAR * Name
        DBCHAR * ActualName
        DBCHAR * TableName
        SHORT Type
        DBINT UserType
        DBINT MaxLength
        BYTE Precision
        BYTE Scale
        BOOL VarLength
        BYTE Null
        BYTE CaseSensitive
        BYTE Updatable
        BOOL Identity

    ctypedef struct            DBDATEREC:
        DBINT year
        DBINT month
        DBINT day
        DBINT dayofyear
        DBINT weekday
        DBINT hour
        DBINT minute
        DBINT second
        DBINT millisecond
        DBINT tzone

    ctypedef struct            DBDATEREC2:
        DBINT year
        DBINT quarter
        DBINT month
        DBINT day
        DBINT dayofyear
        DBINT week
        DBINT weekday
        DBINT hour
        DBINT minute
        DBINT second
        DBINT nanosecond
        DBINT tzone

    ctypedef struct DBNUMERIC:
        BYTE precision
        BYTE scale
        BYTE array[33]

    ctypedef DBNUMERIC DBDECIMAL

    # Error handler callback
    ctypedef int(*EHANDLEFUNC)(DBPROCESS *, int, int, int, char *, char *)

    # Message handler callback
    ctypedef int(*MHANDLEFUNC)(DBPROCESS *, DBINT, int, int, char *, char *, char *, int)

    ## Constants ##
    int FAIL
    int SUCCEED
    int INT_CANCEL
    int NO_MORE_ROWS
    int NO_MORE_RESULTS
    int REG_ROW
    int DBNOERR
    int DBRPCRETURN

    int CI_ALTERNATE
    int CI_CURSOR
    int CI_REGULAR
    int MAXCOLNAMELEN

    ## Bulk Copy Constants ##
    int DB_IN
    int DB_OUT
    int BCPHINTS

    ## Version Constants ##
    int DBVERSION_42
    int DBVERSION_70
    int DBVERSION_71
    int DBVERSION_72
    int DBVERSION_73
    int DBVERSION_74

    ## Type Constants ##
    cdef enum:
        SYBBINARY
        SYBBIT
        SYBCHAR
        SYBDATETIME
        SYBDATETIME4
        SYBDATETIMN
        SYBDECIMAL
        SYBFLT8
        SYBFLTN
        SYBIMAGE
        SYBINT1
        SYBINT2
        SYBINT4
        SYBINT8
        SYBINTN
        SYBMONEY
        SYBMONEY4
        SYBMONEYN
        SYBNUMERIC
        SYBREAL
        SYBTEXT
        SYBVARBINARY
        SYBVARCHAR
        SYBMSDATETIMEOFFSET

    ## Primary functions ##

    # See which version of db-lib is in use.
    #
    #   Returns:
    #     null-terminated ASCII string representing the version of db-lib.
    const char *dbversion()

    # Get address of compute column data.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #     computeid of COMPUTE clause to which we're referring.
    #     column    Nth column in computeid, starting from 1.
    #
    #   Returns:
    #     pointer to columns' data buffer.
    #
    #   Return values:
    #     NULL no such compute id or column.
    BYTE * dbadata(DBPROCESS *, int, int) nogil

    # Get address of compute column data.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #     computeid of COMPUTE clause to which we're referring.
    #     column    Nth column in computeid, starting from 1.
    #
    #   Returns:
    #     size of the data, in bytes
    #
    #   Return values:
    #     -1 no such column or computeid.
    #      0 data is NULL.
    DBINT dbadlen(DBPROCESS *, int, int) nogil

    # Get datatype for a compute column data.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #     computeid of COMPUTE clause to which we're referring.
    #     column    Nth column in computeid, starting from 1.
    #
    #   Returns:
    #     SYB* datatype token
    #
    #   Return values:
    #     -1 no such column or computeid.
    int dbalttype(DBPROCESS *, int, int) nogil

    # Cancel the current command batch
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #
    #   Returns:
    #     SUCCEED always
    RETCODE dbcancel(DBPROCESS *) nogil

    # Close a connection to the server and free associated resources.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    void dbclose(DBPROCESS *) nogil

    # Append SQL to the command buffer.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #     cmdstring SQL to append to the command buffer.
    #
    #   Returns:
    #     SUCCEED   success
    #     FAIL      insufficient memory
    RETCODE dbcmd(DBPROCESS *, char *)

    # Return name of a regular result column.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #     column    Nth in the result set, starting with 1.
    #
    #   Returns:
    #     pointer to ASCII null-terminated string, the name of the column.
    #
    #   Return values:
    #     NULL      column is not in range.
    char * dbcolname(DBPROCESS *, int)

    # Get the datatype of a regular result set column.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #     column    Nth column in computeid, starting from 1.
    #
    #   Returns:
    #     SYB* datatype token value, or zero if column is out of range.
    int dbcoltype(DBPROCESS *, int) nogil

    # Convert one datatype to another.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #     srctype   datatype of the data to convert
    #     src       buffer to convert
    #     srclen    length of src
    #     desttype  target datatype
    #     dest      output buffer
    #     destlen   size of dest
    #
    #   Returns:
    #     On success, the count of output bytes in dest, else -1. On
    #     failure, it will call any user-supplied error handler.
    DBINT dbconvert(DBPROCESS *, int, BYTE *, DBINT, int, BYTE *, DBINT)

    ctypedef struct DBTYPEINFO:
        DBINT precision
        DBINT scale

    DBINT dbconvert_ps(DBPROCESS * dbprocess, int srctype, BYTE * src, DBINT srclen, int desttype, BYTE * dest, DBINT destlen,
        DBTYPEINFO * typeinfo)

    # Get count of rows processed.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #   Returns:
    #     * for insert/update/delete, count of rows affected.
    #     * for select, count of rows returned, after all rows have been
    #       fetched.
    DBINT dbcount(DBPROCESS *) nogil

    # Check if dbproc is an ex-parrot.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #   Returns:
    #     * for insert/update/delete, count of rows affected.
    #     * for select, count of rows returned, after all rows have been
    DBBOOL dbdead(DBPROCESS *)

    # Get address of data in a regular result column.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #     column    Nth column in computeid, starting from 1.
    #
    #   Returns:
    #     pointer to the data, or NULL if data is NULL, or if column is
    #     out of range
    BYTE * dbdata(DBPROCESS *, int) nogil

    # Break any kind of date or time value into useful pieces.
    #
    #   Parameters:
    #     dbproc     contains all information needed by db-lib to manage
    #                     communications with the server.
    #     di             output: structure to contain the exploded parts of
    #                     datetime.
    #     type         input: type of date/time value returned by dbcoltype().
    #     datetime  input: DBDATETIME to be converted.
    #
    #   Return values:
    #     SUCCEED always
    #
    # remarks
    #           The members of \a di have different names, depending on whether \c --with-msdblib was configured.
    #           This is an extension to dbdatecrack(), see it for more information.
    RETCODE dbanydatecrack(DBPROCESS * , DBDATEREC2 * , int , const void *)

    # Get size of current row's data in a regular result column.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #     column    Nth column in computeid, starting from 1.
    #
    #   Returns:
    #     size of the data, in bytes
    DBINT dbdatlen(DBPROCESS *, int) nogil

    # Set an error handler, for messages from db-lib.
    #
    #   Parameters:
    #       handler pointer to callback function that will handle errors.
    #               Pass NULL to restore the default handler.
    #
    #   Returns:
    #       address of prior handler, or NULL if none was previously
    #       installed.
    EHANDLEFUNC dberrhandle(EHANDLEFUNC)

    # Close server connections and free all related structures.
    void dbexit()

    # Get maximum simultaneous connections db-lib will open to the server.
    #
    #   Returns:
    #     size of the data, in bytes
    int dbgetmaxprocs()

    # Initialize db-lib
    #
    #   Return values:
    #     SUCCEED   normal
    #     FAIL      cannot allocate an array of TDS_MAX_CONN TDSSOCKET
    #               pointers.
    RETCODE dbinit()

    # Allocate a LOGINREC structure.
    #
    #   Return values:
    #     NULL      the LOGINREC cannot be allocated.
    #     LOGINREC* to valid memory
    LOGINREC * dblogin()

    # free the LOGINREC
    void dbloginfree(LOGINREC *)

    # Set a message handler, for messages from the server.
    #
    #   Parameters:
    #       handler address of the function that will process the messages.
    MHANDLEFUNC dbmsghandle(MHANDLEFUNC)

    # Get name of current database.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #
    #   Returns:
    #     current database name, as null-terminated ASCII string.
    char * dbname(DBPROCESS *)

    # Read result row into the row buffer and into any bound host variables.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #
    #   Return values:
    #     REG_ROW   regular row has been read.
    #     BUF_FULL  reading the next row would cause the buffer to be
    #               exceeded. No row was read from the server.
    #
    #   Returns:
    #     computeid when a compute row is read.
    RETCODE dbnextrow(DBPROCESS *) nogil

    # Return number of regular columns in a result set.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #
    #   Returns:
    #     number of columns in the result set row.
    int dbnumcols(DBPROCESS *)

    # Form a connection with the server.
    #
    #   Parameters:
    #     login   LOGINREC* carrying the account information
    #     server  name of the dataserver to connect to
    #
    #   Returns:
    #     value pointer on successful login.
    #
    #   Return values:
    #     NULL insufficient memory, unable to connect for any reason
    DBPROCESS * dbopen(LOGINREC *, char *) nogil

    # Set up query results.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #
    #   Return values:
    #     SUCCEED   some result are available.
    #     FAIL      query was not processed successfully by the server.
    #     NO_MORE_RESULTS   query produced no results.
    RETCODE dbresults(DBPROCESS *) nogil

    # Set maximum seconds db-lib waits for a server response to a login
    # attempt.
    #
    #   Parameters:
    #     seconds   New limit for application.
    #
    #   Returns:
    #     SUCCEED always
    RETCODE dbsetlogintime(int)

    # Set maximum simultaneous connections db-lib will open to the server.
    #
    #   Parameters:
    #     maxprocs  Limit for process.
    #
    #   Returns:
    #     SUCCEED always
    RETCODE dbsetmaxprocs(int)

    # Set maximum seconds db-lib waits for a server response to query.
    #
    #   Parameters:
    #     seconds   New limit for application.
    #
    #   Returns:
    #     SUCCEED always
    RETCODE dbsettime(int)

    # Send the SQL command to the server and wait for an answer.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #
    #   Return values:
    #     SUCCEED   query was processed without errors.
    #     FAIL      was returned by dbsqlsend() or dbsqlok()
    RETCODE dbsqlexec(DBPROCESS *) nogil

    # Send the SQL command to the server
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #
    #   Return values:
    #     SUCCEED   query was processed without errors.
    #     FAIL      was returned by dbsqlsend() or dbsqlok()
    RETCODE dbsqlsend(DBPROCESS *) nogil

    # Get file descriptor of the socket used by a DBPROCESS to read data coming
    # from the server.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #
    #   Return values:
    #     An integer file descriptor used by the specified DBPROCESS to read
    #     data coming from the server.
    int dbiordesc(DBPROCESS *) nogil

    # Wait for results of a query from the server.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #
    #   Return values:
    #     SUCCEED   everything worked, fetch results with dbnextresults().
    #     FAIL      SQL syntax error, typically.
    RETCODE dbsqlok(DBPROCESS *) nogil

    # Get the TDS version in use for dbproc.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #
    #   Returns:
    #     a DBTDS* token.
    #
    #   Remarks:
    #     The integer values of the constants are counterintuitive.
    int dbtds(DBPROCESS *)

    # Change current database
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #     name      database to use.
    #
    #   Return values:
    #     SUCCEED query was processed without errors.
    #     FAIL    query was not processed
    RETCODE dbuse(DBPROCESS *, char *) nogil

    ## End Primary functions ##

    ## Remote Procedure functions ##
    # Determine if query generated a return status number
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #   Return values:
    #     TRUE      fetch return status with dbretstatus().
    #     FALSE     no return status
    DBBOOL dbhasretstat(DBPROCESS *)

    # Get count of output parameters filled by a stored procedure
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #   Returns:
    #     How many, possibly zero.
    int dbnumrets(DBPROCESS *)

    # Get value of an output parameter filled by a stored procedure.
    #
    #   Parameters:
    #     dbproc 	contains all information needed by db-lib to manage
    #               communications with the server.
    #     retnum    Nth parameter between 1 and the return value from
    #               dbnumrets()
    #   Returns:
    #     Address of a return parameter value, or NULL if no such retnum.
    BYTE * dbretdata(DBPROCESS *, int) nogil

    # Get size of an output parameter filled by a stored procedure.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #     retnum    Nth parameter between 1 and the return value from
    #               dbnumrets()
    #   Returns:
    #     Size of a return parameter value, or NULL if no such retnum.
    int dbretlen(DBPROCESS *, int) nogil

    # Get name of an output parameter filled by a stored procedure.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #     retnum    Nth parameter between 1 and the return value from
    #               dbnumrets()
    #   Returns:
    #     ASCII null-terminated string, NULL if no such retnum.
    char * dbretname(DBPROCESS *, int) nogil

    # Fetch status value returned by query or remote procedure call.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #   Returns:
    #     The return value of the rpc call
    DBINT dbretstatus(DBPROCESS *) nogil

    # Get datatype of a stored procedure's return parameter.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #     retnum    Nth parameter between 1 and the return value from
    #               dbnumrets()
    #   Returns:
    #     SYB* datatype token, or -1 if retnum is out of range.
    int dbrettype(DBPROCESS *, int) nogil

    # Initialize a remote procedure call.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #     rpcname   name of the stored procedure to be run.
    #     options   Only supported option would be DBRPCRECOMPILE, which causes
    #               the stored procedure to be recompiled before executing.
    #   Return values:
    #     SUCCEED   normal
    #     FAIL      on error
    RETCODE dbrpcinit(DBPROCESS *, char *, DBSMALLINT) nogil

    # Add a parameter to a remote procedure call.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #     paramname literal name of the parameter, according to the stored
    #               procedure (starts with '@'). Optional. If not used,
    #               parameters will be passed in order instead of by name.
    #     status    must be DBRPCRETURN, if this parameter is a return
    #               parameter, else 0.
    #     type      datatype of the value parameter e.g. SYBINT4, SYBCHAR
    #     maxlen    Maximum output size of the parameter's value to be returned
    #               by the stored procedure, usually the size of your host
    #               variable. Fixed-length datatypes take -1 (NULL or not).
    #               Non-OUTPUT parameters also use -1. Use 0 to send a NULL
    #               value for a variable length datatype.
    #     datalen   For variable-length datatypes, the byte size of the data
    #               to be sent, exclusive of any null terminator. For
    #               fixed-length datatypes use -1. To send a NULL value, use 0.
    #     value     Address of your host variable.
    #
    #   Return values:
    #     SUCCEED   normal
    #     FAIL      on error
    RETCODE dbrpcparam(DBPROCESS *, char *, BYTE, int, DBINT, DBINT, BYTE *) nogil

    # Execute the procedure and free associated memory.
    #
    #   Parameters:
    #     dbproc    contains all information needed by db-lib to manage
    #               communications with the server.
    #   Return values:
    #     SUCCEED   normal
    #     FAIL      on error
    RETCODE dbrpcsend(DBPROCESS *) nogil
    ## End Remote Procedure functions ##

    ## Bulk Copy Functions ##
    # Commit a set of rows to the table.
    #
    #   Parameters
    #     dbproc  contains all information needed by db-lib to manage communications with the server.
    #
    #   Returns
    #      Count of rows saved, or -1 on error.
    DBINT bcp_batch (DBPROCESS *) nogil

    # Bind a program host variable to a database column.
    #
    #   Parameters
    #
    #     dbproc	contains all information needed by db-lib to manage communications with the server.
    #     varaddr	address of host variable
    #     prefixlen	length of any prefix found at the beginning of varaddr, in bytes.
    #     Use zero for fixed-length datatypes.
    #     varlen	bytes of data in varaddr. Zero for NULL, -1 for fixed-length datatypes.
    #     terminator	byte sequence that marks the end of the data in varaddr
    #     termlen	length of terminator
    #     vartype	datatype of the host variable
    #     table_column	Nth column, starting at 1, in the table.
    #
    #   Return values:
    #     SUCCEED   normal
    #     FAIL      on error
    RETCODE bcp_bind (DBPROCESS *, BYTE *, int, DBINT, BYTE *, int, int, int) nogil

    # Conclude the transfer of data from program variables.
    #
    #   Parameters
    #     dbproc  contains all information needed by db-lib to manage communications with the server.
    #
    #   Returns
    #     The count of rows saved, or -1 on error.
    DBINT bcp_done (DBPROCESS *) nogil

    # Set "hints" for uploading a file. A FreeTDS-only function.
    #
    #   Parameters
    #     dbproc    contains all information needed by db-lib to manage communications with the server.
    #     option    symbolic constant indicating the option to be set, one of:
    #         BCPLABELED Not implemented.
    #         BCPHINTS The hint to be passed when the bulk-copy begins.
    #     value     The string constant for option a/k/a the hint. One of:
    #         ORDER               The data are ordered in accordance with the table's clustered index.
    #         ROWS_PER_BATCH      The batch size
    #         KILOBYTES_PER_BATCH The approximate number of kilobytes to use for a batch size
    #         TABLOCK             Lock the table
    #         CHECK_CONSTRAINTS   Apply constraints
    #         FIRE_TRIGGERS       Fire any INSERT triggers on the target table
    #     valuelen  The strlen of value.
    RETCODE bcp_options (DBPROCESS *, int, BYTE *, int) nogil

    # Write data in host variables to the table.
    #
    #   Parameters
    #     dbproc  contains all information needed by db-lib to manage communications with the server.
    #
    #   Returns:
    #     SUCCEED   normal
    #     FAIL      on error
    #
    RETCODE bcp_sendrow (DBPROCESS *) nogil

    # Prepare for bulk copy operation on a table.
    #
    #   Parameters
    #     dbproc    contains all information needed by db-lib to manage communications with the server.
    #     tblname   the name of the table receiving or providing the data.
    #     hfile     the data file opposite the table, if any.
    #     errfile   the "error file" captures messages and, if errors are encountered, copies of any rows that could not be written to the table.
    #     direction one of
    #        DB_IN writing to the table
    #        DB_OUT writing to the host file
    #
    #   Returns:
    #     SUCCEED   normal
    #     FAIL      on error
    RETCODE bcp_init (DBPROCESS *, const char *, const char *, const char *, int) nogil
    ## End Bulk Copy Functions ##

    ## Macros ##
    DBBOOL DBDEAD(DBPROCESS *)
    RETCODE DBSETLAPP(LOGINREC *x, char *y)
    RETCODE DBSETLHOST(LOGINREC *x, char *y)
    RETCODE DBSETLPWD(LOGINREC *x, char *y)
    RETCODE DBSETLUSER(LOGINREC *x, char *y)
    RETCODE DBSETLCHARSET(LOGINREC *x, char *y)
    RETCODE DBSETLVERSION(LOGINREC *login, BYTE version)
    RETCODE DBSETLDBNAME(LOGINREC *x, char *y)
    RETCODE DBSETLENCRYPT(LOGINREC *login, int x)
    RETCODE DBSETLREADONLY(LOGINREC *login, int x)


cdef extern from "syberror.h":

    # severity levels from syberror.h

    int EXCOMM

ctypedef int LINE_T



# -*- coding: utf-8 -*-
"""
Test usage in threads.
"""

import sys
import threading
import time
import unittest

import pytest

from pymssql._mssql import MSSQLDatabaseException

from .helpers import mssqlconn, StoredProc


error_sproc = StoredProc(
    "pymssqlErrorThreadTest",
    args=(),
    body="SELECT unknown_column FROM unknown_table")


class _TestingThread(threading.Thread):
    def __init__(self):
        super(_TestingThread, self).__init__()
        self.results = []
        self.exc = None

    def run(self):
        try:
            with mssqlconn() as mssql:
                for i in range(0, 1000):
                    num = mssql.execute_scalar('SELECT %d', (i,))
                    assert num == i
                    self.results.append(num)
        except Exception as exc:
            self.exc = exc


@pytest.mark.mssql_server_required
class _TestingErrorThread(_TestingThread):
    def run(self):
        try:
            with mssqlconn() as mssql:
                mssql.execute_query('SELECT unknown_column')
        except Exception as exc:
            self.exc = exc


@pytest.mark.mssql_server_required
class _SprocTestingErrorThread(_TestingThread):
    def run(self):
        try:
            with mssqlconn() as mssql:
                error_sproc.execute(mssql=mssql)
        except Exception as exc:
            self.exc = exc


@pytest.mark.mssql_server_required
class ThreadedTests(unittest.TestCase):
    def run_threads(self, num, thread_class):
        threads = [thread_class() for _ in range(num)]
        for thread in threads:
            thread.start()

        results = []
        exceptions = []

        while len(threads) > 0:
            sys.stdout.write(".")
            sys.stdout.flush()
            for thread in threads:
                if not thread.is_alive():
                    threads.remove(thread)
                if thread.results:
                    results.append(thread.results)
                if thread.exc:
                    exceptions.append(thread.exc)
            time.sleep(5)

        sys.stdout.write(" ")
        sys.stdout.flush()

        return results, exceptions

    @pytest.mark.slow
    def testThreadedUse(self):
        results, exceptions = self.run_threads(
            num=50,
            thread_class=_TestingThread)
        self.assertEqual(len(exceptions), 0)
        for result in results:
            self.assertEqual(result, list(range(0, 1000)))

    @pytest.mark.slow
    def testErrorThreadedUse(self):
        results, exceptions = self.run_threads(
            num=2,
            thread_class=_TestingErrorThread)
        self.assertEqual(len(exceptions), 2)
        for exc in exceptions:
            self.assertEqual(type(exc), MSSQLDatabaseException)

    @pytest.mark.slow
    def testErrorSprocThreadedUse(self):
        with error_sproc.create():
            results, exceptions = self.run_threads(
                num=5,
                thread_class=_SprocTestingErrorThread)
        self.assertEqual(len(exceptions), 5)
        for exc in exceptions:
            self.assertEqual(type(exc), MSSQLDatabaseException)



# -*- coding: utf-8 -*-
"""
Test error handling.
"""

from datetime import datetime
import unittest

import pytest

from pymssql import _mssql


@pytest.mark.mssql_server_required
class ErrHandleTests(unittest.TestCase):

    def test01DBError(self):
        connection = None
        severity = 8
        dberr = 101
        oserr = 0
        dberrstr = "toblerone1"
        oserrstr = None

        expect = "DB-Lib error message %d, severity %d:\n%s\n" % (
            dberr, severity, dberrstr)
        expect = expect.encode('UTF-8')

        values = _mssql.test_err_handler(
            connection, severity, dberr, oserr, dberrstr, oserrstr)
        self.assertEqual(values[0], 2)
        self.assertEqual(values[1], expect)

    def test02OSError(self):
        connection = None
        # EXCOMM
        severity = 9
        dberr = 102
        oserr = 1001
        dberrstr = "toblerone2"
        oserrstr = "scorpion"

        expect = (
            "DB-Lib error message %d, severity %d:\n%s\n"
            "Net-Lib error during %s (%d)\n" % (
                dberr, severity, dberrstr, oserrstr, oserr))
        expect = expect.encode('UTF-8')

        values = _mssql.test_err_handler(
            connection, severity, dberr, oserr, dberrstr, oserrstr)
        self.assertEqual(values[0], 2)
        self.assertEqual(values[1], expect)

    def test03OSError(self):
        connection = None
        severity = 10
        dberr = 103
        oserr = 1003
        dberrstr = "toblerone3"
        oserrstr = "cabezon"

        expect = (
            "DB-Lib error message %d, severity %d:\n%s\n"
            "Operating System error during %s (%d)\n" % (
                dberr, severity, dberrstr, oserrstr, oserr))
        expect = expect.encode('UTF-8')

        values = _mssql.test_err_handler(
            connection, severity, dberr, oserr, dberrstr, oserrstr)
        self.assertEqual(values[0], 2)
        self.assertEqual(values[1], expect)

    def test04NoError(self):
        connection = None

        # smaller than min error severity, so no output should be generated
        severity = 5
        dberr = 10
        oserr = 4444
        dberrstr = "toblerone4"
        oserrstr = "limpet"

        expect = b""

        values = _mssql.test_err_handler(
            connection, severity, dberr, oserr, dberrstr, oserrstr)
        self.assertEqual(values[0], 2)
        self.assertEqual(values[1], expect)



"""
This type stub file was generated by pyright.
"""

import datetime
from collections.abc import Sequence
from typing import Any
from . import _mssql
from .exceptions import *

__author__: str
__full_version__: str
__version__: str
VERSION: tuple[int, ...]
apilevel: str
threadsafety: int
paramstyle: str
NoParams = _mssql.NoParams

class DBAPIType:
    def __init__(self, value: Any) -> None: ...
    def __cmp__(self, other: Any) -> int: ...
    def __eq__(self, other: object) -> bool: ...

STRING: int
BINARY: int
NUMBER: int
DATETIME: int
DECIMAL: int
Date = datetime.date
Time = datetime.time
Timestamp = datetime.datetime
DateFromTicks: Date
TimeFromTicks: Time
TimestampFromTicks: Timestamp
Binary: bytearray

def row2dict(row: tuple[Any, ...]) -> dict[Any, Any]:
    """Filter dict so it only has string keys; used when as_dict == True"""
    ...

class output:
    object: Any
    value: Any
    def __init__(self, param_type: Any, value: Any | None = None) -> None: ...

class Connection:
    """
    This class represents an MS-SQL database connection.
    """

    as_dict: bool
    autocommit_state: bool
    _conn: _mssql.MSSQLConnection
    arraysize: int
    def __init__(
        self,
        conn: _mssql.MSSQLConnection,
        as_dict: bool,
        autocommit: bool,
        arraysize: int = ...,
    ) -> None: ...
    def __dealloc__(self) -> None: ...
    def autocommit(self, status: bool) -> None:
        """
        Turn autocommit ON or OFF.
        """
        ...

    def __enter__(self) -> Connection: ...
    def __exit__(self, exc_type, exc_value, traceback) -> None: ...
    def close(self) -> None:
        """
        Close the connection to the database. Implicitly rolls back all
        uncommitted transactions.
        """
        ...

    def commit(self) -> None:
        """
        Commit transaction which is currently in progress.
        """
        ...

    def cursor(
        self, as_dict: bool | None = None, arraysize: int | None = None
    ) -> Cursor:
        """
        Return cursor object that can be used to make queries and fetch
        results from the database.
        """
        ...

    def rollback(self) -> None:
        """
        Roll back transaction which is currently in progress.
        """
        ...

    def bulk_copy(
        self,
        table_name: str,
        elements: list,
        column_ids: list[int] | None = None,
        batch_size: int = 1000,
        tablock: bool = False,
        check_constraints: bool = False,
        fire_triggers: bool = False,
    ) -> None: ...

class Cursor:
    """
    This class represents a database cursor, which is used to issue queries
    and fetch results from a database connection.
    """

    connection: Connection
    description: tuple[Any, ...]
    lastrowid: int
    rowcount: int
    returnvalue: object
    rownumber: int
    _source: Connection
    arraysize: int
    def __init__(
        self,
        conn: Connection,
        as_dict: bool,
        arraysize: int = 1,
    ) -> None: ...
    def __iter__(self) -> Cursor:
        """
        Return self to make cursors compatible with Python iteration
        protocol.
        """
        ...

    def __enter__(self) -> Cursor: ...
    def __exit__(self, exc_type, exc_value, traceback) -> None: ...
    def callproc(
        self,
        procname: str,
        parameters: Sequence[str] = ...,
    ) -> tuple[Any, ...]:
        """
        Call a stored procedure with the given name.

        :param procname: The name of the procedure to call
        :type procname: str
        :keyword parameters: The optional parameters for the procedure
        :type parameters: sequence
        """
        ...

    def close(self) -> None:
        """
        Closes the cursor. The cursor is unusable from this point.
        """
        ...

    def execute(self, operation: str, params: object = ...) -> None: ...
    def executemany(
        self,
        operation: str,
        seq_of_parameters: Sequence[str],
        *,
        batch_size: int = -1,
    ) -> None: ...
    def nextset(self) -> int | None: ...
    def getrow(self) -> tuple[Any, ...]:
        """
        Helper method used by fetchone and fetchmany to fetch and handle
        converting the row if as_dict = True.
        """
        ...

    def fetchone(self) -> tuple[Any, ...] | None: ...
    def fetchmany(self, size: int | None = None) -> list[tuple[Any, ...]] | None: ...
    def fetchall(self) -> list[tuple[Any, ...]] | None: ...
    def __next__(self) -> tuple[Any, ...] | None: ...
    def setinputsizes(self, sizes: int | None = None) -> None:
        """
        This method does nothing, as permitted by DB-API specification.
        """
        ...

    def setoutputsize(self, size: int | None = None, column: int = 0) -> None:
        """
        This method does nothing, as permitted by DB-API specification.
        """
        ...

def connect(
    server: str = ".",
    user: str | None = None,
    password: str | None = None,
    database: str = "",
    timeout: int = 0,
    login_timeout: int = 60,
    charset: str = "UTF-8",
    as_dict: bool = False,
    host: str = "",
    appname: str | None = None,
    port: str = "1433",
    encryption: str | None = None,
    read_only: bool = False,
    conn_properties: str | list[str] | None = None,
    autocommit: bool = False,
    tds_version: str | None = None,
    use_datetime2=False,
    arraysize: int = 1,
) -> Connection:
    """
    Constructor for creating a connection to the database. Returns a
    Connection object.

    :param server: database host
    :type server: string
    :param user: database user to connect as. Default value: None.
    :type user: string
    :param password: user's password. Default value: None.
    :type password: string
    :param database: the database to initially connect to
    :type database: string
    :param timeout: query timeout in seconds, default 0 (no timeout)
    :type timeout: int
    :param login_timeout: timeout for connection and login in seconds, default 60
    :type login_timeout: int
    :param charset: character set with which to connect to the database
    :type charset: string
    :keyword as_dict: whether rows should be returned as dictionaries instead of tuples.
    :type as_dict: boolean
    :keyword appname: Set the application name to use for the connection
    :type appname: string
    :keyword port: the TCP port to use to connect to the server
    :type port: string
    :keyword conn_properties: SQL queries to send to the server upon connection
                              establishment. Can be a string or another kind
                              of iterable of strings
    :keyword autocommit: Whether to use default autocommitting mode or not
    :type autocommit: boolean
    :keyword tds_version: TDS protocol version to use.
    :type tds_version: string
    :keyword use_datetime2: Whether to use datetime.datetime conversion compatible with DATETIME2. Default: False.
    :type use_datetime2: bool
    :keyword arraysize:
        This read/write attribute specifies the number of rows to fetch at a time
        with .fetchmany(). It defaults to 1 meaning to fetch a single row at a time.
        Default value: 1.
    :type arraysize: int
    """
    ...

def get_max_connections() -> int:
    """
    Get the maximum number of simultaneous connections pymssql will open
    to the server.
    """
    ...

def set_max_connections(limit: int) -> int:
    """
    Set maximum simultaneous connections db-lib will open to the server.

    :param limit: the connection limit
    :type limit: int
    """
    ...

def get_dbversion() -> str:
    """
    Return string representing the version of db-lib.
    """
    ...

def version_info() -> str:
    """
    Returns string with version information about pymssql, FreeTDS, Python and OS.
    """
    ...



# -*- coding: utf-8 -*-
"""
Test queries.
"""

import pytest

from .helpers import pymssqlconn, drop_table
from pymssql._mssql import substitute_params


@pytest.mark.mssql_server_required
class Test_609:

    table_name = 'testtab'
    table_ddl = f'''
        CREATE TABLE {table_name} (
            int_col int,
            int_col_none int,
            text_col nvarchar(100)
            )
    '''
    N = 10
    table_data = tuple( (x, f"Column {x}") for x in range(N) )

    @classmethod
    def setup_class(cls):
        cls.conn = pymssqlconn(encryption='require')
        drop_table(cls.conn._conn, cls.table_name)
        cls.create_test_table()
        cls.fill_test_table()

    @classmethod
    def teardown_class(cls):
        drop_table(cls.conn._conn, cls.table_name)
        cls.conn.close()

    @classmethod
    def create_test_table(cls):
        with cls.conn.cursor() as c:
            c.execute(cls.table_ddl)

    @classmethod
    def fill_test_table(cls):
        query = f"INSERT INTO {cls.table_name} (int_col, text_col) VALUES (%d, %s);"
        with cls.conn.cursor() as c:
            for vals in cls.table_data:
                c.execute(query, vals)


    def test_text0(self):
        with self.conn.cursor() as c:
            c.execute(f'SELECT * FROM {self.table_name} WHERE text_col=%s', "a")
            rows = c.fetchall()
            assert len(rows) == 0


    @pytest.mark.parametrize('i', range(N))
    def test_int(self, i):
        int_col, text_col = self.table_data[i]
        with self.conn.cursor() as c:
            c.execute(f'SELECT * FROM {self.table_name} WHERE int_col=%s', (int_col, ))
            rows = c.fetchall()
            assert len(rows) == 1
            assert rows[0] == (int_col, None, text_col)


    @pytest.mark.parametrize('i', range(N))
    def test_int_pymssql(self, i):
        int_col, text_col = self.table_data[i]
        with self.conn.cursor() as c:
            c.execute(f'SELECT * FROM {self.table_name} WHERE int_col=%s', int_col)
            rows = c.fetchall()
            assert len(rows) == 1
            assert rows[0] == (int_col, None, text_col)


    @pytest.mark.parametrize('i', range(N))
    def test_text(self, i):
        int_col, text_col = self.table_data[i]
        with self.conn.cursor() as c:
            c.execute(f'SELECT * FROM {self.table_name} WHERE text_col=%s', (text_col, ))
            rows = c.fetchall()
            assert len(rows) == 1
            assert rows[0] == (int_col, None, text_col)


    @pytest.mark.parametrize('i', range(N))
    def test_text_pymssql(self, i):
        int_col, text_col = self.table_data[i]
        with self.conn.cursor() as c:
            c.execute(f'SELECT * FROM {self.table_name} WHERE text_col=%s', text_col)
            rows = c.fetchall()
            assert len(rows) == 1
            assert rows[0] == (int_col, None, text_col)


    @pytest.mark.xfail(reason="known parser issue")
    def test_424(self):
        query = "SELECT COUNT(*) FROM document WHERE title LIKE '%summary%' AND id < %s"
        parms = (1500,)
        with self.conn.cursor() as c:
            c.execute(query, parms)


    @pytest.mark.xfail(reason="known parser issue")
    def test_276(self):
        res = substitute_params('select %s;'.encode('utf-8'), tuple(["Фрязино".encode('utf-16-le')]))
        assert res == b'select 0x240440044f04370438043d043e042000;'
        #assert res == b"select '$\x04@\x04O\x047\x048\x04=\x04>\x04';"

        res = substitute_params('select %s;'.encode('utf-8'), tuple(["Фрязино ".encode('utf-16-le')]))
        assert res == b'select 0x240440044f04370438043d043e042000;'
        "Фрязино".encode('utf-16-le')
        '$\x04@\x04O\x047\x048\x04=\x04>\x04'
        "Фрязино ".encode('utf-16-le')
        '$\x04@\x04O\x047\x048\x04=\x04>\x04 \x00'



# -*- coding: utf-8 -*-

import time
import pytest
import pymssql


@pytest.mark.slow
@pytest.mark.mssql_server_required
@pytest.mark.timeout(120)
@pytest.mark.xfail(strict=False)
@pytest.mark.parametrize('to', [2])
def test_remote_connect_timeout(to):

    t = time.time()
    try:
        pymssql.connect(server="www.google.com", port=81, user='username', password='password',
                            login_timeout=to)
    except pymssql.OperationalError:
        pass
    t = time.time() - t
    print('remote: requested {} -> {} actual timeout'.format(to, t))
    assert t == pytest.approx(to, 5), "{} != {}".format(t, to)



"""DB-SIG compliant module for communicating with MS SQL servers"""
# pymssql.pyx
#
#   Copyright (C) 2003 Joon-cheol Park <jooncheol@gmail.com>
#                 2008 Andrzej Kukula <akukula@gmail.com>
#                 2009 Damien Churchill <damoxc@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301  USA

import datetime
import time

from pymssql import _mssql
from pymssql cimport _mssql
from cpython cimport bool
from libc.string cimport strlen
from .sqlfront cimport BYTE

cdef extern from "version.h":
    const char *PYMSSQL_VERSION

__author__ = 'Damien Churchill <damoxc@gmail.com>'
__full_version__ = PYMSSQL_VERSION.decode('ascii')
__version__ = '.'.join(__full_version__.split('.')[:3])
VERSION = tuple(int(c) if c.isdigit() else c for c in __full_version__.split('.')[:3])

# Strives for compliance with DB-API 2.0 (PEP 249)
# http://www.python.org/dev/peps/pep-0249/
apilevel = '2.0'

# module may be shared, but not connections
threadsafety = 1

# this module uses extended python format codes
paramstyle = 'pyformat'

# Singleton for parameterless queries
NoParams = _mssql.NoParams

from pymssql._mssql import set_wait_callback
from pymssql.exceptions import *

# store a tuple of programming error codes
cdef object prog_errors = (
    102,    # syntax error
    207,    # invalid column name
    208,    # invalid object name
    2812,   # unknown procedure
    4104    # multi-part identifier could not be bound
)

# store a tuple of integrity error codes
cdef object integrity_errors = (
    515,    # NULL insert
    547,    # FK related
    2601,   # violate unique index
    2627,   # violate UNIQUE KEY constraint
)

class DBAPIType:

    def __init__(self, value):
        self.value = value

    def __cmp__(self, other):
        if other == self.value:
            return 0
        if other < self.value:
            return 1
        else:
            return -1

    def __eq__(self, other):
        return (other == self.value)

    def __repr__(self):
        return '<DBAPIType %s>' % self.value

STRING = DBAPIType(_mssql.STRING)
BINARY = DBAPIType(_mssql.BINARY)
NUMBER = DBAPIType(_mssql.NUMBER)
DATETIME = DBAPIType(_mssql.DATETIME)
DECIMAL = DBAPIType(_mssql.DECIMAL)

Date = datetime.date
Time = datetime.time
Timestamp = datetime.datetime
DateFromTicks = lambda ticks: Date(*time.localtime(ticks)[:3])
TimeFromTicks = lambda ticks: Time(*time.localtime(ticks)[3:6])
TimestampFromTicks = lambda ticks: Timestamp(*time.localtime(ticks)[:6])
Binary = bytearray


# Bulk copy hints
cdef char* TABLOCK = "TABLOCK"
cdef char* CHECK_CONSTRAINTS = "CHECK_CONSTRAINTS"
cdef char* FIRE_TRIGGERS = "FIRE_TRIGGERS"

cdef int TABLOCK_LEN = strlen(TABLOCK)
cdef int CHECK_CONSTRAINTS_LEN = strlen(CHECK_CONSTRAINTS)
cdef int FIRE_TRIGGERS_LEN = strlen(FIRE_TRIGGERS)


def row2dict(row):
    """Filter dict so it only has string keys; used when as_dict == True"""
    return dict([(k, v) for k, v in row.items() if hasattr(k, 'startswith')])

# stored procedure output parameter
cdef class output:

    cdef object _type
    cdef object _value

    property type:
        """
        This is the type of the parameter.
        """
        def __get__(self):
            return self._type

    property value:
        """
        This is the value of the parameter.
        """
        def __get__(self):
            return self._value


    def __init__(self, param_type, value=None):
        self._type = param_type
        self._value = value

######################
## Connection class ##
######################
cdef class Connection:
    """
    This class represents an MS-SQL database connection.
    """

    cdef bool _as_dict
    cdef bool _autocommit
    cdef _mssql.MSSQLConnection conn
    cdef int _arraysize

    property as_dict:
        """
        Instructs all cursors this connection creates to return results
        as a dictionary rather than a tuple.
        """
        def __get__(self):
            return self._as_dict
        def __set__(self, value):
            self._as_dict = value

    property autocommit_state:
        """
        The current state of autocommit on the connection.
        """
        def __get__(self):
            return self._autocommit

    property _conn:
        """
        INTERNAL PROPERTY. Returns the _mssql.MSSQLConnection object, and
        raise exception if it's set to None. It's easier than adding the
        necessary checks to every other method.
        """
        def __get__(self):
            if self.conn == None:
                raise InterfaceError('Connection is closed.')
            return self.conn

    property arraysize:
        """
        This read/write attribute specifies the number of rows to fetch at a time
        with .fetchmany(). It defaults to 1 meaning to fetch a single row at a time.
        """
        def __get__(self):
            return self._arraysize

        def __set__(self, value):
            cdef int val = int(value)
            if val < 0:
                raise ValueError("The 'arraysize' attribute must be >= 0.")
            self._arraysize = val

    def __init__(self, conn, as_dict, autocommit, arraysize=1):
        self.conn = conn
        self._autocommit = autocommit
        self.as_dict = as_dict
        self.arraysize = arraysize

        if not autocommit:
            try:
                self._conn.execute_non_query('BEGIN TRAN')
            except Exception, e:
                raise OperationalError('Cannot start transaction: ' + str(e.args[0]))

    def __dealloc__(self):
        if self.conn:
            self.close()

    def autocommit(self, status):
        """
        Turn autocommit ON or OFF.
        """

        if status == self._autocommit:
            return

        tran_type = 'ROLLBACK' if status else 'BEGIN'
        self._conn.execute_non_query('%s TRAN' % tran_type)

        self._autocommit = status

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        """
        Close the connection to the database. Implicitly rolls back all
        uncommitted transactions.
        """
        if self.conn:
            self.conn.close()
        self.conn = None

    def commit(self):
        """
        Commit transaction which is currently in progress.
        """

        if self._autocommit == True:
            return

        try:
            self._conn.execute_non_query('COMMIT TRAN')
            self._conn.execute_non_query('BEGIN TRAN')
        except Exception, e:
            raise OperationalError('Cannot commit transaction: ' + str(e.args[0]))

    def cursor(self, as_dict=None, arraysize=None):
        """
        Return cursor object that can be used to make queries and fetch
        results from the database.
        """
        if as_dict is None:
            as_dict = self.as_dict
        return Cursor(self, as_dict, arraysize or self.arraysize)

    def rollback(self):
        """
        Roll back transaction which is currently in progress.
        """
        if self._autocommit == True:
            return

        try:
            self._conn.execute_non_query('ROLLBACK TRAN')
        except _mssql.MSSQLException, e:
            # PEP 249 indicates that we have contract with the user that we will
            # always have a transaction in place if autocommit is False.
            # Therefore, it seems logical to ignore this exception since it
            # indicates a situation we shouldn't ever encounter anyway.  However,
            # it can happen when an error is severe enough to cause a
            # "batch-abort".  In that case, SQL Server *implicitly* rolls back
            # the transaction for us (how helpful!).  But there doesn't seem
            # to be any way for us to know if an error is severe enough to cause
            # a batch abort:
            #   http://stackoverflow.com/questions/5877162/why-does-microsoft-sql-server-implicitly-rollback-when-a-create-statement-fails
            #
            # the alternative is to do 'select @@trancount' before each rollback
            # but that is slower and doesn't seem to offer any benefit.
            if 'The ROLLBACK TRANSACTION request has no corresponding BEGIN TRANSACTION' not in str(e):
                raise
        try:
            self._conn.execute_non_query('BEGIN TRAN')
        except Exception, e:
            raise OperationalError('Cannot begin transaction: ' + str(e.args[0]))

    cpdef bulk_copy(self,
                    table_name,
                    elements,
                    column_ids=None,
                    batch_size=1000,
                    tablock=False,
                    check_constraints=False,
                    fire_triggers=False):
        cdef int batch_counter = 0
        cdef _mssql.MSSQLConnection conn = self._conn
        conn.bcp_init(table_name)

        if tablock:
            conn.bcp_hint(<BYTE*> TABLOCK, TABLOCK_LEN)
        if check_constraints:
            conn.bcp_hint(<BYTE*> CHECK_CONSTRAINTS, CHECK_CONSTRAINTS_LEN)
        if fire_triggers:
            conn.bcp_hint(<BYTE*> FIRE_TRIGGERS, FIRE_TRIGGERS_LEN)

        for element in elements:
            if batch_counter == batch_size:
                conn.bcp_batch()
                batch_counter = 0

            self._conn.bcp_sendrow(element, column_ids)
            batch_counter += 1
        conn.bcp_done()


##################
## Cursor class ##
##################
cdef class Cursor:
    """
    This class represents a database cursor, which is used to issue queries
    and fetch results from a database connection.
    """

    cdef Connection conn
    cdef public tuple description
    cdef int _rownumber
    cdef bool as_dict
    cdef object _returnvalue
    cdef int _arraysize

    property connection:
        def __get__(self):
            return self.conn

    property lastrowid:
        def __get__(self):
            return self.conn.conn.identity

    property rowcount:
        def __get__(self):
            return self._rownumber

    property returnvalue:
        def __get__(self):
            return self._returnvalue

    property rownumber:
        def __get__(self):
            return self._rownumber

    property _source:
        def __get__(self):
            if self.conn == None:
                raise InterfaceError('Cursor is closed.')
            return self.conn

    property arraysize:
        """
        This read/write attribute specifies the number of rows to fetch at a time
        with .fetchmany(). It defaults to 1 meaning to fetch a single row at a time.
        """
        def __get__(self):
            return self._arraysize

        def __set__(self, value):
            cdef int val = int(value)
            if val < 0:
                raise ValueError("The 'arraysize' attribute must be >= 0.")
            self._arraysize = val

    def __init__(self, conn, as_dict, arraysize=1):
        self.conn = conn
        self.description = None
        self._rownumber = 0
        self._returnvalue = None
        self.as_dict = as_dict
        self.arraysize = arraysize

    def __iter__(self):
        """
        Return self to make cursors compatible with Python iteration
        protocol.
        """
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def callproc(self, str procname, parameters=()):
        """
        Call a stored procedure with the given name.

        :param procname: The name of the procedure to call
        :type procname: str
        :keyword parameters: The optional parameters for the procedure
        :type parameters: sequence
        """
        self._returnvalue = None
        proc = self._source._conn.init_procedure(procname)
        for parameter in parameters:
            if type(parameter) is output:
                param_type = parameter.type
                param_value = parameter.value
                param_output = True
            else:
                param_type = type(parameter)
                param_value = parameter
                param_output = False

            db_type = _mssql.py2db_type(param_type, param_value)

            proc.bind(param_value, db_type, output=param_output)
        try:
            self._returnvalue = proc.execute()
            self.description = self._source._conn.get_header()
        except _mssql.MSSQLDatabaseException, e:
            raise DatabaseError, e.args[0]
        return tuple([proc.parameters[p] for p in proc.parameters])

    def close(self):
        """
        Closes the cursor. The cursor is unusable from this point.
        """
        self.conn = None
        self.description = None

    def execute(self, operation, params=NoParams):
        self.description = None
        self._rownumber = 0

        try:
            self._source._conn.execute_query(operation, params)
            self.description = self._source._conn.get_header()
            self._rownumber = self._source._conn.rows_affected

            if self.as_dict and self.description:
                columns_without_names = [
                    idx
                    for idx, column_descriptor in enumerate(self.description)
                    if len(column_descriptor[0]) == 0
                ]
                if columns_without_names:
                    raise ColumnsWithoutNamesError(columns_without_names)

        except _mssql.MSSQLDatabaseException, e:
            if e.number in prog_errors:
                raise ProgrammingError, e.args[0]
            if e.number in integrity_errors:
                raise IntegrityError, e.args[0]
            raise OperationalError, e.args[0]
        except _mssql.MSSQLDriverException, e:
            raise InterfaceError, e.args[0]

    def executemany(self, operation, seq_of_parameters, *, batch_size=-1):
        self.description = None
        self._rownumber = -1
        if batch_size == -1:
            batch_size = self.arraysize
        if batch_size == 1:
            rownumber = 0
            for params in seq_of_parameters:
                self.execute(operation, params)
                # support correct rowcount across multiple executes
                rownumber += self._rownumber
            self._rownumber = rownumber
        else:
            self._source._conn.executemany(operation, seq_of_parameters, batch_size)

    def nextset(self):
        try:
            if not self._source._conn.nextresult():
                return None
            self._rownumber = 0
            self.description = self._source._conn.get_header()
            return 1

        except _mssql.MSSQLDatabaseException, e:
            raise OperationalError, e.args[0]
        except _mssql.MSSQLDriverException, e:
            raise InterfaceError, e.args[0]

    cdef getrow(self):
        """
        Helper method used by fetchone and fetchmany to fetch and handle
        converting the row if as_dict = True.
        """
        row_format = _mssql.ROW_FORMAT_DICT if self.as_dict else _mssql.ROW_FORMAT_TUPLE
        row = next(self._source._conn.get_iterator(row_format))
        if not self.as_dict:
            return row
        return row2dict(row)

    def fetchone(self):
        if self.description is None:
            raise OperationalError('Statement not executed or executed statement has no resultset')

        try:
            return self.getrow()
        except StopIteration:
            self._rownumber = self._source._conn.rows_affected
            return None
        except _mssql.MSSQLDatabaseException, e:
            raise OperationalError, e.args[0]
        except _mssql.MSSQLDriverException, e:
            raise InterfaceError, e.args[0]

    def fetchmany(self, size=None):
        if self.description is None:
            raise OperationalError('Statement not executed or executed statement has no resultset')

        if size == None:
            size = self.arraysize
        else:
            self.arraysize = size

        try:
            rows = []
            for i in xrange(size):
                try:
                    rows.append(self.getrow())
                except StopIteration:
                    self._rownumber = self._source._conn.rows_affected
                    break
            return rows
        except _mssql.MSSQLDatabaseException, e:
            raise OperationalError, e.args[0]
        except _mssql.MSSQLDriverException, e:
            raise InterfaceError, e.args[0]

    def fetchall(self):
        if self.description is None:
            raise OperationalError('Statement not executed or executed statement has no resultset')

        try:
            rows = []
            while True:
                try:
                    rows.append(self.getrow())
                except StopIteration:
                    break
            self._rownumber = self._source._conn.rows_affected
            return rows
        except _mssql.MSSQLDatabaseException, e:
            raise OperationalError, e.args[0]
        except _mssql.MSSQLDriverException, e:
            raise InterfaceError, e.args[0]

    def __next__(self):
        try:
            row = self.getrow()
            self._rownumber += 1
            return row

        except _mssql.MSSQLDatabaseException, e:
            raise OperationalError, e.args[0]
        except _mssql.MSSQLDriverException, e:
            raise InterfaceError, e.args[0]

    def setinputsizes(self, sizes=None):
        """
        This method does nothing, as permitted by DB-API specification.
        """
        pass

    def setoutputsize(self, size=None, column=0):
        """
        This method does nothing, as permitted by DB-API specification.
        """
        pass

def connect(server='.', user=None, password=None, database='', timeout=0,
        login_timeout=60, charset='UTF-8', as_dict=False,
        host='', appname=None, port='1433', encryption=None, read_only=False,
        conn_properties=None, autocommit=False, tds_version=None,
        use_datetime2=False,
        arraysize=1):
    """
    Constructor for creating a connection to the database. Returns a
    Connection object.

    :param server: database host
    :type server: string
    :param user: database user to connect as. Default value: None.
    :type user: string
    :param password: user's password. Default value: None.
    :type password: string
    :param database: the database to initially connect to
    :type database: string
    :param timeout: query timeout in seconds, default 0 (no timeout)
    :type timeout: int
    :param login_timeout: timeout for connection and login in seconds, default 60
    :type login_timeout: int
    :param charset: character set with which to connect to the database
    :type charset: string
    :keyword as_dict: whether rows should be returned as dictionaries instead of tuples.
    :type as_dict: boolean
    :keyword appname: Set the application name to use for the connection
    :type appname: string
    :keyword port: the TCP port to use to connect to the server
    :type port: string
    :keyword conn_properties: SQL queries to send to the server upon connection
                              establishment. Can be a string or another kind
                              of iterable of strings
    :keyword autocommit: Whether to use default autocommitting mode or not
    :type autocommit: boolean
    :keyword tds_version: TDS protocol version to use.
    :type tds_version: string
    :keyword use_datetime2: Whether to use datetime.datetime conversion compatible with DATETIME2
    :type use_datetime2: bool
    :keyword arraysize:
        This read/write attribute specifies the number of rows to fetch at a time
        with .fetchmany(). It defaults to 1 meaning to fetch a single row at a time.
        Default value: 1.
    :type arraysize: int
    """

    # set the login timeout
    try:
        login_timeout = int(login_timeout)
    except ValueError:
        login_timeout = 0

    _mssql.login_timeout = login_timeout

    # default query timeout
    try:
        timeout = int(timeout)
    except ValueError:
        timeout = 0

    if host:
        server = host

    try:
        conn = _mssql.connect(server=server, encryption=encryption,
                              user=user, password=password,
                              charset=charset, database=database,
                              appname=appname, port=port, tds_version=tds_version,
                              read_only=read_only,
                              use_datetime2=use_datetime2,
                              conn_properties=conn_properties)

    except _mssql.MSSQLDatabaseException, e:
        raise OperationalError(e.args[0])

    except _mssql.MSSQLDriverException, e:
        raise InterfaceError(e.args[0])


    if timeout != 0:
        conn.query_timeout = timeout

    return Connection(conn, as_dict, autocommit, arraysize=arraysize)

def get_max_connections():
    """
    Get the maximum number of simultaneous connections pymssql will open
    to the server.
    """
    return _mssql.get_max_connections()

def set_max_connections(int limit):
    """
    Set maximum simultaneous connections db-lib will open to the server.

    :param limit: the connection limit
    :type limit: int
    """
    _mssql.set_max_connections(limit)

def get_dbversion():
    """
    Return string representing the version of db-lib.
    """
    return _mssql.get_dbversion()

get_freetds_version = get_dbversion

def version_info():
    """
        Returns string with version information about pymssql, FreeTDS, Python and OS.
    """

    import platform
    import sys
    import io

    output = io.StringIO()
    print("System:", file=output)
    print("platform.system()       :", platform.system(), file=output)
    print("platform.architecture() :", platform.architecture(), file=output)
    if platform.system() != 'Windows':
        print("platform.libc_ver()     :", platform.libc_ver(), file=output)

    print("Python:", file=output)
    print("sys.version             :", sys.version, file=output)

    print("pymssql:", file=output)
    print("VERSION            :", VERSION, file=output)
    print("version            :", __version__, file=output)
    print("full_version       :", __full_version__, file=output)
    print("dbversion          :", get_dbversion(), file=output)

    print("_mssql:", file=output)
    print("VERSION            :", _mssql.VERSION, file=output)
    print("version            :", _mssql.__version__, file=output)
    print("full_version       :", _mssql.__full_version__, file=output)
    print("dbversion          :", _mssql.get_dbversion(), file=output)
    print("login_timeout      :", _mssql.login_timeout, file=output)
    print("min_error_severity :", _mssql.min_error_severity, file=output)
    contents = output.getvalue()
    output.close()
    return contents





# exception hierarchy
class Warning(Exception):
    pass

class Error(Exception):
    pass

class InterfaceError(Error):
    pass

class DatabaseError(Error):
    pass

class DataError(DatabaseError):
    pass

class OperationalError(DatabaseError):
    pass

class IntegrityError(DatabaseError):
    pass

class InternalError(DatabaseError):
    pass

class ProgrammingError(DatabaseError):
    pass

class NotSupportedError(DatabaseError):
    pass

class ColumnsWithoutNamesError(InterfaceError):
    def __init__(self, columns_without_names):
        self.columns_without_names = columns_without_names

    def __str__(self):
        return 'Specified as_dict=True and ' \
            'there are columns with no names: %r' \
            % (self.columns_without_names,)



# -*- coding: utf-8 -*-
"""
Test stored procedure usage.
"""

from contextlib import contextmanager
import decimal
import datetime
import os
import unittest

import pymssql
from pymssql import _mssql

import pytest

from .helpers import mssqlconn, pymssqlconn


@contextmanager
def stored_proc(mssqlcon, dbtype):

    dbtype = dbtype.lower()
    identifier = dbtype
    if dbtype == 'decimal':
        identifier = 'decimal(6, 5)'
    elif dbtype == 'numeric':
        identifier = 'numeric(6, 5)'
    elif dbtype == 'time':
        identifier = 'time(7)'
    elif dbtype == 'datetimeoffset':
        identifier = 'datetimeoffset(7)'

    name = f"pymssqlTest_{dbtype}"

    cmd = f"""
        CREATE PROCEDURE [dbo].[{name}]
            @inp {identifier},
            @out {identifier} output
        AS
        BEGIN
            SET @out = @inp;
            RETURN 0;
        END
    """
    mssqlcon.execute_non_query(cmd)

    try:
        yield name
    finally:
        # Code to release resource, e.g.:
        mssqlcon.execute_non_query(f'DROP PROCEDURE [dbo].[{name}]')


FIXED_TYPES = (
    'BigInt',
    'Bit',
    'Date',
    'DateTime',
    'Decimal',
    'Int',
    'Money',
    'Numeric',
    'SmallInt',
    'TinyInt',
    'UniqueIdentifier',
    'Real',
    'Float'
)

VARIABLE_TYPES = (
    ('Char', 4),
    ('VarChar', 4),
    ('VarBinary', 4),
    ('Text', None)  # Leave this one in the last position in case it fails (see https://code.google.com/p/pymssql/issues/detail?id=113#c2)
)

@pytest.mark.mssql_server_required
class TestFixedTypeConversion(unittest.TestCase):

    def setUp(self):
        self.mssql = mssqlconn()
        self.pymssql = pymssqlconn()

        for name in FIXED_TYPES:
            dbtype = name.lower()
            if dbtype == 'decimal':
                identifier = 'decimal(6, 5)'
            elif dbtype == 'numeric':
                identifier = 'numeric(6, 5)'
            else:
                identifier = dbtype

            self.mssql.execute_non_query("""
            CREATE PROCEDURE [dbo].[pymssqlTest%(name)s]
                @i%(dbtype)s %(identifier)s,
                @o%(dbtype)s %(identifier)s output
            AS
            BEGIN
                SET @o%(dbtype)s = @i%(dbtype)s;
                RETURN 0;
            END
            """ % {
                'dbtype': dbtype,
                'name': name,
                'identifier': identifier
            })

    def tearDown(self):
        for name in FIXED_TYPES:
            self.mssql.execute_non_query('DROP PROCEDURE [dbo].[pymssqlTest%s]' % name)

        self.pymssql.close()
        self.mssql.close()

    def testBigIntNullIn(self):
        input = None
        proc = self.mssql.init_procedure('pymssqlTestBigInt')
        proc.bind(input, _mssql.SQLINT8, '@ibigint', null=True)
        proc.bind(None, _mssql.SQLINT8, '@obigint', output=True)
        proc.execute()
        self.assertEqual(input, proc.parameters['@obigint'])

    def testBigInt(self):
        input = 123456789
        proc = self.mssql.init_procedure('pymssqlTestBigInt')
        proc.bind(input, _mssql.SQLINT8, '@ibigint')
        proc.bind(None, _mssql.SQLINT8, '@obigint', output=True)
        proc.execute()
        self.assertEqual(input, proc.parameters['@obigint'])

    def testBigIntPymssql(self):
        """Same as testBigInt above but from pymssql. Uses pymssql.output class."""
        in_val = 123456789
        cursor = self.pymssql.cursor()
        retval = cursor.callproc('pymssqlTestBigInt', [in_val, pymssql.output(int)])
        self.assertEqual(in_val, retval[1])

        in_val = 2147483647
        retval = cursor.callproc('pymssqlTestBigInt', [in_val, pymssql.output(int)])
        self.assertEqual(in_val, retval[1])

    def testBit(self):
        input = True
        proc = self.mssql.init_procedure('pymssqlTestBit')
        proc.bind(input, _mssql.SQLBITN, '@ibit')
        proc.bind(False, _mssql.SQLBITN, '@obit', output=True)
        proc.execute()
        self.assertEqual(input, proc.parameters['@obit'])

    def testDateTime(self):
        input = datetime.datetime(2009, 8, 27, 15, 28, 38)
        proc = self.mssql.init_procedure('pymssqlTestDateTime')
        proc.bind(input, _mssql.SQLDATETIME, '@idatetime')
        proc.bind(None, _mssql.SQLDATETIME, '@odatetime', output=True)
        proc.execute()
        self.assertEqual(input, proc.parameters['@odatetime'])

    def testDecimal(self):
        input = decimal.Decimal('5.12345')
        output = decimal.Decimal('0.00000')
        proc = self.mssql.init_procedure('pymssqlTestDecimal')
        proc.bind(input, _mssql.SQLDECIMAL, '@idecimal')
        proc.bind(output, _mssql.SQLDECIMAL, '@odecimal', output=True, max_length=6)
        proc.execute()
        self.assertEqual(input, proc.parameters['@odecimal'])
        self.assertEqual(str(input), str(proc.parameters['@odecimal']))

    def testDecimal2(self):
        input = decimal.Decimal('6.23456')
        output = decimal.Decimal('0.00000')
        proc = self.mssql.init_procedure('pymssqlTestDecimal')
        proc.bind(input, _mssql.SQLDECIMAL, '@idecimal')
        proc.bind(output, _mssql.SQLDECIMAL, '@odecimal', output=True, max_length=6)
        proc.execute()
        self.assertEqual(input, proc.parameters['@odecimal'])
        self.assertEqual(str(input), str(proc.parameters['@odecimal']))

    def testDecimal3(self):
        output = decimal.Decimal('0.00000')
        proc = self.mssql.init_procedure('pymssqlTestDecimal')
        input = decimal.Decimal('6.23400')
        proc.bind(input, _mssql.SQLDECIMAL, '@idecimal')
        proc.bind(output, _mssql.SQLDECIMAL, '@odecimal', output=True, max_length=6)
        proc.execute()
        self.assertEqual(input, proc.parameters['@odecimal'])
        self.assertEqual(str(input), str(proc.parameters['@odecimal']))

    def testDecimal4(self):
        output = decimal.Decimal('1.0000000')
        proc = self.mssql.init_procedure('pymssqlTestDecimal')
        input = decimal.Decimal('6.2340000')
        proc.bind(input, _mssql.SQLDECIMAL, '@idecimal')
        proc.bind(output, _mssql.SQLDECIMAL, '@odecimal', output=True, max_length=15)
        proc.execute()
        self.assertEqual(input, proc.parameters['@odecimal'])
        self.assertEqual(str(input), str(proc.parameters['@odecimal']))

    def testDecimal5(self):
        output = decimal.Decimal('1.000000000')
        proc = self.mssql.init_procedure('pymssqlTestDecimal')
        input = decimal.Decimal('6.234000000')
        proc.bind(input, _mssql.SQLDECIMAL, '@idecimal')
        proc.bind(output, _mssql.SQLDECIMAL, '@odecimal', output=True, max_length=15)
        proc.execute()
        self.assertEqual(input, proc.parameters['@odecimal'])
        self.assertEqual(str(input), str(proc.parameters['@odecimal']))

    def testInt(self):
        input = 10056
        proc = self.mssql.init_procedure('pymssqlTestInt')
        proc.bind(input, _mssql.SQLINT4, '@iint')
        proc.bind(None, _mssql.SQLINT4, '@oint', output=True)
        proc.execute()
        self.assertEqual(input, proc.parameters['@oint'])

    def testMoney(self):
        input = decimal.Decimal('5.12')
        proc = self.mssql.init_procedure('pymssqlTestMoney')
        proc.bind(input, _mssql.SQLMONEY, '@imoney')
        proc.bind(None, _mssql.SQLMONEY, '@omoney', output=True)
        proc.execute()
        self.assertEqual(input, proc.parameters['@omoney'])

    def testNumeric(self):
        input = decimal.Decimal('5.12345')
        output = decimal.Decimal('0.00000')
        proc = self.mssql.init_procedure('pymssqlTestNumeric')
        proc.bind(input, _mssql.SQLNUMERIC, '@inumeric')
        proc.bind(output, _mssql.SQLNUMERIC, '@onumeric', output=True)
        proc.execute()
        self.assertEqual(input, proc.parameters['@onumeric'])
        self.assertEqual(str(input), str(proc.parameters['@onumeric']))

    def testSmallInt(self):
        input = 10056
        proc = self.mssql.init_procedure('pymssqlTestSmallInt')
        proc.bind(input, _mssql.SQLINT2, '@ismallint')
        proc.bind(None, _mssql.SQLINT2, '@osmallint', output=True)
        proc.execute()
        self.assertEqual(input, proc.parameters['@osmallint'])

    def testTinyInt(self):
        input = 101
        proc = self.mssql.init_procedure('pymssqlTestTinyInt')
        proc.bind(input, _mssql.SQLINT1, '@itinyint')
        proc.bind(None, _mssql.SQLINT1, '@otinyint', output=True)
        proc.execute()
        self.assertEqual(input, proc.parameters['@otinyint'])

    def testUuid(self):
        if os.environ.get('FREETDS_VERSION') != '0.91':
            pytest.skip("UNIQUEIDENTIFIER as a SP param doesn't work with FreeTDS >= 0.95")
        import uuid
        input = uuid.uuid4()
        proc = self.mssql.init_procedure('pymssqlTestUniqueIdentifier')
        proc.bind(input, _mssql.SQLUUID, '@iuniqueidentifier')
        proc.bind(None, _mssql.SQLUUID, '@ouniqueidentifier', output=True)
        proc.execute()
        self.assertEqual(input, proc.parameters['@ouniqueidentifier'])

    def testReal(self):
        input = 3.14
        proc = self.mssql.init_procedure('pymssqlTestReal')
        proc.bind(input, _mssql.SQLREAL, '@ireal')
        proc.bind(None, _mssql.SQLREAL, '@oreal', output=True)
        proc.execute()
        assert abs(input - proc.parameters['@oreal']) < 0.00001

    def testFloat8(self):
        input = 3.40E38 + 1
        proc = self.mssql.init_procedure('pymssqlTestFloat')
        proc.bind(input, _mssql.SQLFLT8, '@ifloat')
        proc.bind(None, _mssql.SQLFLT8, '@ofloat', output=True)
        proc.execute()
        assert abs(input - proc.parameters['@ofloat']) < 0.00001


@pytest.mark.mssql_server_required
class TestDateTimeConversion:

    def test_date(self, mssql_conn, subtests):
        with stored_proc(mssql_conn, 'date') as proc_name:
            for input in (
                datetime.date(1, 1, 1), # min supported
                datetime.date(90, 8, 27), # GH #454
                datetime.date(906, 8, 27), # GH #454
                datetime.date(2009, 8, 27),
                datetime.date(9999, 12, 31), # max supported
                ):
                with subtests.test(date=input):
                    proc = mssql_conn.init_procedure(proc_name)
                    proc.bind(input, _mssql.SQLDATE, '@inp')
                    proc.bind(None, _mssql.SQLDATE, '@out', output=True)
                    res = proc.execute()
                    assert res == 0
                    assert input == proc.parameters['@out']

    def test_time(self, mssql_conn, subtests):
        with stored_proc(mssql_conn, 'time') as proc_name:
            for input in (
                datetime.time(0, 0, 0), # min supported
                datetime.time(23, 59, 59, 0),
                datetime.time(23, 59, 59, 999999), # max supported
                ):
                with subtests.test(time=input):
                    proc = mssql_conn.init_procedure(proc_name)
                    proc.bind(input, _mssql.SQLTIME, '@inp')
                    proc.bind(None, _mssql.SQLTIME, '@out', output=True)
                    res = proc.execute()
                    assert res == 0
                    assert input == proc.parameters['@out']

    def test_datetime(self, mssql_conn, subtests):
        with stored_proc(mssql_conn, 'datetime') as proc_name:
            for input in (
                datetime.datetime(1753, 1, 1), # min supported
                datetime.datetime(2009, 8, 27),
                datetime.datetime(2009, 8, 27, 23, 59, 59, 997000),
                datetime.datetime(9999, 12, 31, 23, 59, 59, 997000), # max supported
                ):
                with subtests.test(datetime=input):
                    proc = mssql_conn.init_procedure(proc_name)
                    proc.bind(input, _mssql.SQLDATETIME, '@inp')
                    proc.bind(None, _mssql.SQLDATETIME, '@out', output=True)
                    res = proc.execute()
                    assert res == 0
                    assert input == proc.parameters['@out']

    def test_datetime2(self, mssql_conn, subtests):
        with stored_proc(mssql_conn, 'datetime2') as proc_name:
            for input in (
                datetime.datetime(1, 1, 1), # min supported
                datetime.datetime(90, 8, 27),
                datetime.datetime(906, 8, 27),
                datetime.datetime(2009, 8, 27),
                datetime.datetime(2009, 8, 27, 23, 59, 59, 999999),
                datetime.datetime(9999, 12, 31, 23, 59, 59, 999999), # max supported
                ):
                with subtests.test(datetime=input):
                    proc = mssql_conn.init_procedure(proc_name)
                    proc.bind(input, _mssql.SQLDATETIME2, '@inp')
                    proc.bind(None, _mssql.SQLDATETIME2, '@out', output=True)
                    res = proc.execute()
                    assert res == 0
                    assert input == proc.parameters['@out']

    def test_datetimeoffset(self, mssql_conn, subtests):
        with stored_proc(mssql_conn, 'datetimeoffset') as proc_name:
            for h in range(-12,13,1):
                for m in range(0, 61, 15):
                    tz = datetime.timezone(datetime.timedelta(hours=h, minutes=m))
                    for input in (
                        datetime.datetime(10, 1, 2, tzinfo=tz), # min supported
                        datetime.datetime(2999, 12, 31, 23, 59, 59, 999999, tzinfo=tz), # max supported
                        ):
                        with subtests.test(datetime=input):
                            proc = mssql_conn.init_procedure(proc_name)
                            proc.bind(input, _mssql.SQLDATETIMEOFFSET, '@inp')
                            proc.bind(None, _mssql.SQLDATETIMEOFFSET, '@out', output=True)
                            res = proc.execute()
                            assert res == 0
                            assert input == proc.parameters['@out']


@pytest.mark.mssql_server_required
class TestCallProcFancy(unittest.TestCase):
    """
    "Fancy" because we test some exotic cases like passing None or Unicode
    # strings to a called procedure
    """

    def setUp(self):
        self.pymssql = pymssqlconn()
        cursor = self.pymssql.cursor()

        sql = """
        CREATE PROCEDURE [dbo].[someProcWithOneParam]
            @some_arg NVARCHAR(64)
        AS
        BEGIN
            SELECT @some_arg + N'!',
                   N'%(str1)s ' + @some_arg + N' %(str2)s'
        END
        """ % {
            'str1': 'Здравствуй',
            'str2': 'Мир',
        }
        sql = sql.encode('utf-8')

        cursor.execute(sql)

    def tearDown(self):
        cursor = self.pymssql.cursor()
        cursor.execute('DROP PROCEDURE [dbo].[someProcWithOneParam]')
        self.pymssql.close()

    def testCallProcWithNone(self):
        cursor = self.pymssql.cursor()
        cursor.callproc(
            'someProcWithOneParam',
            (None,))

        a, b = cursor.fetchone()
        self.assertEqual(a, None)
        self.assertEqual(b, None)

    def testCallProcWithAsciiString(self):
        cursor = self.pymssql.cursor()
        cursor.callproc(
            'someProcWithOneParam',
            ('hello',))

        a, b = cursor.fetchone()
        self.assertEqual(a, 'hello!')
        self.assertEqual(b, 'Здравствуй hello Мир')

    def testCallProcWithUnicodeStringWithNoFunnyCharacters(self):
        cursor = self.pymssql.cursor()
        cursor.callproc(
            'someProcWithOneParam',
            ('hello',))

        a, b = cursor.fetchone()
        self.assertEqual(a, 'hello!')
        self.assertEqual(b, 'Здравствуй hello Мир')

    # This is failing for me - the Unicode params somehow gets rendered to a
    # blank string. I am not sure if this is another bug or a user error on my
    # part...?
    #
    def testCallProcWithUnicodeStringWithRussianCharacters(self):
        cursor = self.pymssql.cursor()
        cursor.callproc(
            'someProcWithOneParam',
            ('Здравствуй',))  # Russian string

        a, b = cursor.fetchone()
        self.assertEqual(a, 'Здравствуй!')
        self.assertEqual(b, 'Здравствуй Здравствуй Мир')

    def testExecuteWithNone(self):
        cursor = self.pymssql.cursor()
        cursor.execute(
            'someProcWithOneParam %s',
            (None,))

        a, b = cursor.fetchone()
        self.assertEqual(a, None)
        self.assertEqual(b, None)

    def testExecuteWithAsciiString(self):
        cursor = self.pymssql.cursor()
        cursor.execute(
            'someProcWithOneParam %s',
            ('hello',))

        a, b = cursor.fetchone()
        self.assertEqual(a, 'hello!')
        self.assertEqual(b, 'Здравствуй hello Мир')

    def testExecuteWithUnicodeStringWithNoFunnyCharacters(self):
        cursor = self.pymssql.cursor()
        cursor.execute(
            'someProcWithOneParam %s',
            ('hello',))

        a, b = cursor.fetchone()
        self.assertEqual(a, 'hello!')
        self.assertEqual(b, 'Здравствуй hello Мир')

    def testExecuteWithUnicodeWithRussianCharacters(self):
        cursor = self.pymssql.cursor()
        cursor.execute(
            'someProcWithOneParam %s',
            ('Здравствуй',))  # Russian string

        a, b = cursor.fetchone()
        self.assertEqual(a, 'Здравствуй!')
        self.assertEqual(b, 'Здравствуй Здравствуй Мир')



@pytest.mark.mssql_server_required
class TestStringTypeConversion(unittest.TestCase):

    def setUp(self):
        self.mssql = mssqlconn()

        for name, size in VARIABLE_TYPES:
            dbtype = name.lower()
            identifier = dbtype if dbtype == 'text' else '%s(%d)' % (dbtype, size)

            try:
                self.mssql.execute_non_query("""
                CREATE PROCEDURE [dbo].[pymssqlTest%(name)s]
                    @i%(dbtype)s %(identifier)s,
                    @o%(dbtype)s %(identifier)s output
                AS
                BEGIN
                    SET @o%(dbtype)s = @i%(dbtype)s;
                    RETURN 0;
                END
                """ % {
                    'dbtype': dbtype,
                    'name': name,
                    'identifier': identifier
                })
            except:
                if name == 'Text':
                    raise

    def tearDown(self):
        for name, size in VARIABLE_TYPES:
            self.mssql.execute_non_query('DROP PROCEDURE [dbo].[pymssqlTest%s]' % name)
        self.mssql.close()

    def testChar(self):
        input = 'test'
        proc = self.mssql.init_procedure('pymssqlTestChar')
        proc.bind(input, _mssql.SQLCHAR, '@ichar')
        proc.bind(None, _mssql.SQLCHAR, '@ochar', output=True, max_length=4)
        proc.execute()
        self.assertEqual(input, proc.parameters['@ochar'])

    def testText(self):
        input = 'test'
        proc = self.mssql.init_procedure('pymssqlTestText')
        proc.bind(input, _mssql.SQLTEXT, '@itext')
        proc.bind(None, _mssql.SQLVARCHAR, '@otext', output=True)
        proc.execute()
        self.assertEqual(input, proc.parameters['@otext'])

    def testVarChar(self):
        input = 'test'
        proc = self.mssql.init_procedure('pymssqlTestVarChar')
        proc.bind(input, _mssql.SQLVARCHAR, '@ivarchar')
        proc.bind(None, _mssql.SQLVARCHAR, '@ovarchar', output=True)
        proc.execute()
        self.assertEqual(input, proc.parameters['@ovarchar'])

    def testVarBinary(self):
        def check_conversion(input, output_type):
            proc = self.mssql.init_procedure('pymssqlTestVarBinary')
            proc.bind(input, _mssql.SQLVARBINARY, '@ivarbinary')
            proc.bind(None, _mssql.SQLVARBINARY, '@ovarbinary', output=True)
            proc.execute()
            self.assertEqual(input, proc.parameters['@ovarbinary'])
            self.assertEqual(output_type, type(proc.parameters['@ovarbinary']))

        check_conversion(bytes(b'\xDE\xAD\xBE\xEF'), bytes)
        check_conversion(bytearray(b'\xDE\xAD\xBE\xEF'), bytes)
        with pytest.raises(TypeError) as exc_info:
            check_conversion('FOO', bytes)
            assert 'value can only be bytes or bytearray' == str(exc_info.value)


@pytest.mark.mssql_server_required
class TestFloatTypeConversion(unittest.TestCase):

    def setUp(self):
        self.mssql = mssqlconn()
        self.pymssql = pymssqlconn()
        self.cursor = self.pymssql.cursor()

        self.mssql.execute_non_query("""
            CREATE PROCEDURE [dbo].[pymssqlRealTest]
                @inparam real
            AS
            BEGIN
                SELECT @inparam AS outparam
            END
            """
        )

        self.mssql.execute_non_query("""
            CREATE PROCEDURE [dbo].[pymssqlFloatTest]
                @inparam float
            AS
            BEGIN
                SELECT @inparam AS outparam
            END
            """
        )

    def tearDown(self):
        self.mssql.execute_non_query('DROP PROCEDURE [dbo].[pymssqlRealTest]')
        self.mssql.execute_non_query('DROP PROCEDURE [dbo].[pymssqlFloatTest]')
        self.mssql.close()
        self.cursor.close()
        self.pymssql.close()

    def test_real(self):
        self.cursor.callproc('pymssqlRealTest', (0.5,))
        a = next(self.cursor)
        assert len(a) == 1
        assert abs(a[0] - 0.5) < 0.000001

    def test_float8(self):
        self.cursor.callproc('pymssqlFloatTest', (5.44451787074e+39,))
        a = next(self.cursor)
        assert len(a) == 1
        assert abs(a[0] - 5.44451787074e+39) < 0.000001


@pytest.mark.mssql_server_required
class TestFetch(unittest.TestCase):

    def setUp(self):
        self.pymssql = pymssqlconn()
        self.cursor = self.pymssql.cursor()

        self.cursor.execute("""
            CREATE PROCEDURE [dbo].[pymssqlTest]
                @inparam real
            AS
            BEGIN
                SELECT @inparam AS outparam
            END
            """
        )

    def tearDown(self):
        self.cursor.execute('DROP PROCEDURE [dbo].[pymssqlTest]')
        self.cursor.close()
        self.pymssql.close()

    def test_fetchall(self):
        self.cursor.callproc('pymssqlTest', (0.5,))
        a = self.cursor.fetchall()
        assert len(a) == 1
        assert len(a[0]) == 1
        assert abs(a[0][0] - 0.5) < 0.000001

    def test_fetchone(self):
        self.cursor.callproc('pymssqlTest', (0.5,))
        a = self.cursor.fetchone()
        assert len(a) == 1
        assert abs(a[0] - 0.5) < 0.000001

    def test_fetchmany(self):
        self.cursor.callproc('pymssqlTest', (0.5,))
        a = self.cursor.fetchmany()
        assert len(a) == 1
        assert len(a[0]) == 1
        assert abs(a[0][0] - 0.5) < 0.000001

    def test_real_next(self):
        self.cursor.callproc('pymssqlTest', (0.5,))
        a = next(self.cursor)
        assert len(a) == 1
        assert abs(a[0] - 0.5) < 0.000001


@pytest.mark.mssql_server_required
class TestErrorInSP(unittest.TestCase):

    def setUp(self):
        self.pymssql = pymssqlconn()
        self.cursor = self.pymssql.cursor()

        sql = """
        CREATE PROCEDURE [dbo].[SPThatRaisesAnError]
        AS
        BEGIN
            -- RAISERROR -- Generates an error message and initiates error processing for the session.
            -- http://msdn.microsoft.com/en-us/library/ms178592.aspx
            -- Severity levels from 0 through 18 can be specified by any user.
            RAISERROR('Error message', 18, 1)
            RETURN
        END
        """
        self.cursor.execute(sql)

    def tearDown(self):
        self.cursor.execute('DROP PROCEDURE [dbo].[SPThatRaisesAnError]')
        self.cursor.close()
        self.pymssql.close()

    def test_tsql_to_python_exception_translation(self):
        """
        An error raised by a SP is translated to a PEP-249-dictated,
        pymssql layer exception.
        See https://github.com/pymssql/pymssql/issues/61
        """
        # Must raise an exception
        self.assertRaises(Exception, self.cursor.callproc, 'SPThatRaisesAnError')
        # Must be a PEP-249 exception, not a _mssql-layer one
        try:
            self.cursor.callproc('SPThatRaisesAnError')
        except Exception as e:
            self.assertTrue(isinstance(e,  pymssql.Error))
        # Must be a DatabaseError exception
        try:
            self.cursor.callproc('SPThatRaisesAnError')
        except Exception as e:
            self.assertTrue(isinstance(e,  pymssql.DatabaseError))


@pytest.mark.mssql_server_required
class TestSPWithQueryResult(unittest.TestCase):

    SP_NAME = 'SPWithAQuery'

    def setUp(self):
        self.mssql = mssqlconn()
        self.pymssql = pymssqlconn()

        self.mssql.execute_non_query("""
        CREATE PROCEDURE [dbo].[%(spname)s]
                @some_arg NVARCHAR(64)
        AS
        BEGIN
                SELECT @some_arg + N'!', @some_arg + N'!!'
        END
        """ % {'spname': self.SP_NAME})

    def tearDown(self):
        self.mssql.execute_non_query('DROP PROCEDURE [dbo].[%(spname)s]' % {'spname': self.SP_NAME})
        self.pymssql.close()
        self.mssql.close()

    def test_pymssql(self):
        cursor = self.pymssql.cursor()
        cursor.callproc(
            self.SP_NAME,
            ('hello',))

        a, b = cursor.fetchone()
        self.assertEqual(a, 'hello!')
        self.assertEqual(b, 'hello!!')

    def test_mssql(self):
        proc = self.mssql.init_procedure(self.SP_NAME)
        proc.bind('hello', _mssql.SQLVARCHAR)
        proc.execute()

        for row_dict in self.mssql:
            self.assertEqual(row_dict, {0: 'hello!', 1: 'hello!!'})


@pytest.mark.mssql_server_required
class TestMultipleDataSetsInResult(unittest.TestCase):
    """
    Test for https://github.com/pymssql/pymssql/pull/134
    See also https://gist.github.com/msabramo/6747703
    """
    table_name = 'callproc_demo_tbl'
    proc_name = 'callproc_get_2_resultsets'

    def setUp(self):
        self.pymssql = pymssqlconn()
        self.cursor = self.pymssql.cursor()
        self.cursor.execute(f"IF OBJECT_ID('{self.table_name}', 'U') IS NOT NULL DROP TABLE {self.table_name}")
        self.cursor.execute(f"CREATE TABLE {self.table_name} (name VARCHAR(30))")
        self.cursor.execute(f"INSERT INTO {self.table_name} VALUES ('Tom'), ('Dick'), ('Harry')")
        self.cursor.execute(f"""
        CREATE PROCEDURE [dbo].[{self.proc_name}] AS
        BEGIN
            SET NOCOUNT ON
            SELECT * FROM {self.table_name} WHERE name = 'not here';
            SELECT * FROM {self.table_name};
        END
        """)

    def tearDown(self):
        self.cursor.execute(f'DROP PROCEDURE [dbo].[{self.proc_name}]')
        self.cursor.close()
        self.pymssql.close()

    def test_exec(self):
        self.cursor.execute(f'EXEC {self.proc_name}')
        res = self.cursor.fetchall()
        self.assertEqual(res, [])
        res = self.cursor.fetchall()
        self.assertEqual(res, [('Tom',), ('Dick',), ('Harry',)])

    @pytest.mark.xfail(strict=True)
    def test_callproc(self):
        self.cursor.callproc(self.proc_name)
        res = self.cursor.fetchall()
        self.assertEqual(res, [])
        res = self.cursor.fetchall()
        self.assertEqual(res, [('Tom',), ('Dick',), ('Harry',)])


@pytest.mark.mssql_server_required
class TestSPWithNULLResult(unittest.TestCase):
    """
    GH: #441
    """

    SP_NAME = 'SPWithANullResult'

    def setUp(self):
        self.mssql = mssqlconn()
        self.pymssql = pymssqlconn()

        self.mssql.execute_non_query("""
        CREATE PROCEDURE [dbo].[%(spname)s]
                @int_arg int output,
                @varchar_arg varchar output
        AS
        BEGIN
                SET @int_arg = null
                SET @varchar_arg = null
        END
        """ % {'spname': self.SP_NAME})

    def tearDown(self):
        self.mssql.execute_non_query('DROP PROCEDURE [dbo].[%(spname)s]' % {'spname': self.SP_NAME})
        self.pymssql.close()
        self.mssql.close()

    def testNull(self):
        proc = self.mssql.init_procedure(self.SP_NAME)
        proc.bind(None, _mssql.SQLINT8, '@int_arg', output=True)
        proc.bind(None, _mssql.SQLVARCHAR, '@varchar_arg', output=True)
        proc.execute()
        self.assertIsNone(proc.parameters['@int_arg'])
        self.assertIsNone(proc.parameters['@varchar_arg'])

